from PyQt5.QtWidgets import QWidget, QVBoxLayout
from PyQt5 import QtGui

from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

# Prettier math font
import matplotlib as mpl
mpl.rcParams['mathtext.fontset'] = 'cm'

class MathTextLabel(QWidget):

    def __init__(self, mathText, parent=None, **kwargs):
        super(QWidget, self).__init__(parent, **kwargs)

        l = QVBoxLayout(self)
        l.setContentsMargins(0, 0, 0, 0)

        r, g, b, a = self.palette().base().color().getRgbF()

        self._figure = Figure(edgecolor=(r, g, b), facecolor=(r, g, b))
        self._figure.clear()
        text = self._figure.suptitle(
            mathText,
            x=0.0,
            y=1.0,
            horizontalalignment='left',
            verticalalignment='top',
            size=QtGui.QFont().pointSize()*1.5
        )
        self._canvas = FigureCanvas(self._figure)
        self._canvas.draw()
        l.addWidget(self._canvas)

        (x0, y0), (x1, y1) = text.get_window_extent().get_points()
        w = int(x1-x0)
        h = int(y1-y0)

        self._figure.set_size_inches(w/80, h/80)
        self.setFixedSize(w, h)


'''
class for particle detection
'''

import numpy as np
import scipy as sp
import matplotlib.pyplot as plt

from skimage.feature import peak_local_max # find local max on the image


class Detectors(object):
    """
    detectors class for particle detection and subpixel localisation
    """
    def __init__(self):
        """Initialise variables
        """

        self.img_sef=[]
        self.binary_sef=[]
        
        # parameters for approach
        #SEF
        self.c=0.8 # coef for the thresholding
        self.sigma=3. # max sigma for LOG     
        
        #thresholding
        self.min_distance=4 # minimum distance between two max after SEF
        self.threshold_rel=0.1 # min picl value in relation to the image
        
        self.expected_size=20 # expected size of the particle
        
    def sef(self, img, sigma, c):   
        '''
        spot enhancing filter
        '''
        img_filtered=img*np.ones(img.shape)
        img_sef1=sp.ndimage.gaussian_laplace(img_filtered, sigma) # calculate laplacian of gaussian
        img_sef1=abs(img_sef1-np.abs(np.max(img_sef1))) # remove negative values keeping the proportion b/w pixels

    # thresholding
        th=np.mean(img_sef1)+c*np.std(img_sef1) # calculate threshold value
        img_sef=np.copy(img_sef1) # copy the image
        img_sef[img_sef<th]=0 # thresholding 
    
        # create a binary mask for the next step
        img_sef_bin = np.copy(img_sef)
        img_sef_bin[img_sef_bin<th]=0
        img_sef_bin[img_sef_bin>=th]=1
    
        return img_sef, img_sef_bin
    
    def radialsym_centre(self, img):
        '''
         Calculates the center of a 2D intensity distribution (calculation of radial symmetry centers)  
    
        '''
        
        def lsradialcenterfit(m, b, w):
            '''
            least squares solution to determine the radial symmetry center
            '''
            wm2p1=np.divide(w,(np.multiply(m,m)+1))
            sw=np.sum(wm2p1)
            smmw = np.sum(np.multiply(np.multiply(m,m),wm2p1))
            smw  = np.sum(np.multiply(m,wm2p1))
            smbw = np.sum(np.multiply(np.multiply(m,b),wm2p1))
            sbw  = np.sum(np.multiply(b,wm2p1))
            det = smw*smw - smmw*sw
            xc = (smbw*sw - smw*sbw)/det # relative to image center
            yc = (smbw*smw - smmw*sbw)/det # relative to image center
                
            return xc, yc
    
        # GRID
        #  number of grid points
        Ny, Nx = img.shape
        
        # for x
        val=int((Nx-1)/2.0-0.5)
        xm_onerow = np.asarray(range(-val,val+1))
        xm = np.ones((Nx-1,Nx-1))*xm_onerow
        
        # for y
        val=int((Ny-1)/2.0-0.5)
        ym_onerow = np.asarray(range(-val,val+1))
        ym = (np.ones((Ny-1,Ny-1))*ym_onerow).transpose()
    
        # derivate along 45-degree shidted coordinates
    
        dIdu = np.subtract(img[0:Nx-1, 1:Ny].astype(float),img[1:Nx, 0:Ny-1].astype(float))
        dIdv = np.subtract(img[0:Nx-1, 0:Ny-1].astype(float),img[1:Nx, 1:Ny].astype(float))
        
        
        #smoothing
        filter_core=np.ones((3,3))/9
        fdu=sp.signal.convolve2d(dIdu,filter_core,  mode='same', boundary='fill', fillvalue=0)
        fdv=sp.signal.convolve2d(dIdv,filter_core,  mode='same', boundary='fill', fillvalue=0)
    
        dImag2=np.multiply(fdu,fdu)+np.multiply(fdv,fdv)
    
        #slope of the gradient
        m = np.divide(-(fdv + fdu), (fdu-fdv))
        
        # if some of values in m is NaN 
        m[np.isnan(m)]=np.divide(dIdv+dIdu, dIdu-dIdv)[np.isnan(m)]
        
        # if some of values in m is still NaN
        m[np.isnan(m)]=0 
        
        
        # if some of values in m  are inifinite
        
        m[np.isinf(m)]=10*np.max(m)
        
        #shortband b
        b = ym - m*xm
        
        #weighting
        sdI2=np.sum(dImag2)
        
        xcentroid = np.sum(np.multiply(dImag2, xm))/sdI2
        ycentroid = np.sum(np.multiply(dImag2, ym))/sdI2
        w=np.divide(dImag2, np.sqrt(np.multiply((xm-xcentroid),(xm-xcentroid))+np.multiply((ym-ycentroid),(ym-ycentroid))))
        
        # least square minimisation
        xc,yc=lsradialcenterfit(m, b, w)
        
        # output replated to upper left coordinate
        x=xc + (Nx+1)/2 # xc + (Nx+1)/2
        y=yc + (Ny+1)/2 # yc + (Ny+1)/2
        
        return x, y

    def detect(self, frame):
        '''
        detect vesicles
        '''

        # Spot enhancing filter   
        self.img_sef, self.binary_sef=self.sef(frame, self.sigma, self.c)
        
        # find local maximum
        peaks_coor=peak_local_max(self.img_sef, min_distance=self.min_distance, threshold_rel=self.threshold_rel) # min distance between peaks and threshold_rel - min value of the peak - in relation to the max value


        coordinates=[]  
        for point in peaks_coor:
            
            data=np.zeros((self.expected_size,self.expected_size))
            
            #start point
            start_x=int(point[0]-self.expected_size/2)
            start_y=int(point[1]-self.expected_size/2)
            
            #end point
            end_x=int(point[0]+self.expected_size/2)
            end_y=int(point[1]+self.expected_size/2)
            
            x_0=0
            x_1=self.expected_size
            y_0=0
            y_1=self.expected_size
            
            # define ROI coordinates
            
            if start_x<0:
                start_x=0
                x_0=int(self.expected_size/2-point[0])
                
            if start_y<0:
                start_y=0
                y_0=int(self.expected_size/2-point[1]) 
                
            if end_x>frame.shape[0]:
                end_x=frame.shape[0]
                x_1=int(frame.shape[0]-point[0]+self.expected_size/2)

            if end_y>frame.shape[1]:
                end_y=frame.shape[1]
                y_1=int(frame.shape[1]-point[1]+self.expected_size/2)
            
            
            data[x_0:x_1,y_0:y_1]=frame[start_x:end_x, start_y:end_y]
            
            # radial symmetry centers
            x,y=self.radialsym_centre(data)
            
            # check that the centre is inside of the spot            
            if y<self.expected_size and x<self.expected_size and y>=0 and x>=0:               
                coordinates.append([x+int(point[0]-self.expected_size/2),y+int(point[1]-self.expected_size/2)])

        return coordinates



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

# Models used for ADC and SD analysis
class ModelBase:
    def __init__(self):
        self.R = 0.0
        self.dt = 0.0

class ModelBrownian(ModelBase):
    """Model for free, unrestricted diffusion.
    """
    lower = [0.0, 0.0]
    upper = np.inf
    initial = [0.5e-12, 2.0e-9]

    def __call__(self, t, D, delta):
        return D + delta**2 / (2*t*(1-2*self.R*self.dt/t))

class ModelConfined(ModelBase):
    """Model for confined diffusion.
    """
    lower = [0.0, 0.0, 0.0]
    upper = np.inf
    initial = [0.5e-12, 2.0e-9, 1.0e-3]

    def __call__(self, t, D_micro, delta, tau):
        return D_micro * (tau/t) * (1 - np.exp(-t/tau)) + \
            delta ** 2 / (2 * t * (1 - 2 * self.R * self.dt / t))

class ModelHop(ModelBase):
    """Model for hop diffusion.
    """
    lower = [0.0, 0.0, 0.0, 0.0]
    upper = np.inf
    initial = [0.5e-12, 0.5e-12, 2.0e-9, 1.0e-3]

    def __call__(self, t, D_macro, D_micro, delta, tau):
        return D_macro + \
            D_micro * (tau/t) * (1 - np.exp(-t/tau)) + \
            delta ** 2 / (2 * t * (1 - 2 * self.R * self.dt / t))

class ModelImmobile(ModelBase):
    """Model for immobile diffusion.
    """
    upper = [np.inf]
    lower = [0.0]
    initial = [0.5e-12]

    def __call__(self, t, delta):
        return delta**2 / (2*t*(1-2*self.R*self.dt/t))

class ModelHopModified(ModelBase):
    """Modified model for hop diffusion.
    """
    lower = [0.0, 0.0, 0.0, 0.0]
    upper = np.inf
    initial = [0.5e-12, 0.5e-12, 0.0, 1.0e-3]
    def __call__(self, t, D_macro, D_micro, alpha, tau):
        return alpha * D_macro + \
            (1.0 - alpha) * D_micro * (1 - np.exp(-t/tau))


# Models used for MSD analysis
class ModelLinear(ModelBase):
    """Linear model for MSD analysis."""
    def __call__(self, t, D, delta2):
        # 4 * D * t + 2 * delta2 - 8 * D * R * dt
        return 4.0 * D * (t - 2.0 * self.R * self.dt) + 2.0 * delta2 

class ModelPower(ModelBase):
    """Generic power law model for MSD analysis."""
    def __call__(self, t, D, delta2, alpha):
        # 4 * D * t**alpha + 2 * delta2 - 8 * D * R * dt
        return 4.0 * D * (t**alpha - 2.0 * self.R * self.dt) + 2.0 * delta2 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Class for detection linking
'''


import numpy as np
from scipy.optimize import linear_sum_assignment

class Track(object):

    def __init__(self, first_point, first_frame, trackIdCount):

        self.track_id = trackIdCount  # track ID
        self.trace_frame = [first_frame]  # frame
        self.skipped_frames = 0  # number of frames skipped undetected
        self.trace = [first_point]  # trace path


class Tracker(object):
    """
    links detections frame to frame
    
    """

    def __init__(self, dist_thresh=30, max_frames_to_skip=20, trackIdCount=0):

        self.dist_thresh = dist_thresh
        self.max_frames_to_skip = max_frames_to_skip
        self.tracks = []
        self.trackIdCount = trackIdCount
        self.completeTracks=[]
        
        print(" - - - - - tracker: - - - - - - - ")
        print("dist_thresh ", dist_thresh)
        print("max_frames_to_skip ", max_frames_to_skip)
        
    def cost_calculation(self, detections):
        '''
        calculates cost matrix based on the distance
        '''
        N = len(self.tracks)
        M = len(detections)
        cost = np.zeros((N, M))   # Cost matrix
        for i in range(len(self.tracks)):
            for j in range(len(detections)):
                try:
                    diff = np.array(self.tracks[i].trace[len(self.tracks[i].trace)-1]) - np.array(detections[j])
                    distance = np.sqrt((diff[0])**2 +(diff[1])**2 )

                    cost[i][j] = distance
                except:
                    pass
        cost_array=np.asarray(cost)
        cost_array[cost_array>self.dist_thresh]=10000
        cost=cost_array.tolist()
        return cost
    
    def assignDetectionToTracks(self, cost):
        '''
        assignment based on Hungerian Algorithm
        '''

        N = len(self.tracks)
        assignment = []
        for _ in range(N):
            assignment.append(-1)
        row_ind, col_ind = linear_sum_assignment(cost)
        for i in range(len(row_ind)):
            assignment[row_ind[i]] = col_ind[i]
        
        return assignment
    
    
    def update(self, detections, frameN):
        '''
        main linking function
        '''

        # create tracks if no tracks  found
        if (len(self.tracks) == 0):
            for i in range(len(detections)):
                track = Track(detections[i], frameN, self.trackIdCount)
                self.trackIdCount += 1
                self.tracks.append(track)

        # tracking the targets if there were tracks before
        else:
            
            # Calculate cost using sum of square distance between predicted vs detected centroids
            cost=self.cost_calculation(detections)

            # assigning detection to tracks
            assignment=self.assignDetectionToTracks(cost)

 
            # add the position to the assigned tracks and detect annasigned tracks
            un_assigned_tracks = []
            
            for i in range(len(assignment)):
                if (assignment[i] != -1):
                    # check with the cost distance threshold and unassign if cost is high
                    if (cost[i][assignment[i]] > self.dist_thresh):
                        assignment[i] = -1
                        un_assigned_tracks.append(i)
                        self.tracks[i].skipped_frames += 1

                    else: # add the detection to the track
                        self.tracks[i].trace.append(detections[assignment[i]])
                        self.tracks[i].trace_frame.append(frameN)
                        self.tracks[i].skipped_frames =0
                        
                else:
                    un_assigned_tracks.append(i)
                    self.tracks[i].skipped_frames += 1                

                        
            # Unnasigned detections
            un_assigned_detects = []
            for i_det in range(len(detections)):
                    if i_det not in assignment:
                        un_assigned_detects.append(i_det)
    
            # Start new tracks
            if(len(un_assigned_detects) != 0):
                for i in range(len(un_assigned_detects)):
                    track = Track(detections[un_assigned_detects[i]], frameN,
                                  self.trackIdCount)              


                    self.trackIdCount += 1
                    self.tracks.append(track)
                        
                    
            del_tracks = [] # list of tracks to delete
            
        #remove tracks which have too many skipped frames
            for i in range(len(self.tracks)):
                if (self.tracks[i].skipped_frames > self.max_frames_to_skip):
                    del_tracks.append(i)        
        
        
        
        # delete track
     
            if len(del_tracks) > 0:   

                val_compensate_for_del=0
                for id in del_tracks:
                    new_id=id-val_compensate_for_del

                    self.completeTracks.append(self.tracks[new_id])
                    del self.tracks[new_id]
                    val_compensate_for_del+=1


                      
        print("number of detections: ", len(detections))
        print("number of tracks: ", len(self.tracks))
        print() 




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import tqdm
import warnings
import logging
import csv
from scipy import optimize
from scipy import interpolate
import pandas as pd

from trait2d.exceptions import *

import itertools
import os
from concurrent.futures import ProcessPoolExecutor

class Borg:
    _shared_state = {}
    def __init__(self):
        self.__dict__ = self._shared_state

class ModelDB(Borg):
    """Singleton class holding all models that should be used in analysis."""
    models = []
    def __init__(self):
        Borg.__init__(self)
    def add_model(self, model):
        """Add a new model class to the ModelDB.

        Parameters
        ----------
        model:
            Model class (*not* an instance) to add. There are predefined models available in
            trait2d.analysis.models. Example usage:

            .. code-block:: python

                from trait2d.analysis.models import ModelConfined
                ModelDB().add_model(ModelConfined)
        """
        for m in self.models:
            if m.__class__ == model:
                raise ValueError("ModelDB already contains an instance of the model {}.".format(model.__name__))
        self.models.append(model())

    def get_model(self, model):
        """
        Return the model instance from ModelDB.

        Parameters
        ----------
        model:
            Model class (*not* and instance) to remove. Example usage:

            .. code-block:: python
            
                from trait2d.analysis.models import ModelConfined
                ModelDB().get_model(ModelConfined).initial = [1.0e-12, 1.0e-9, 0.5e-3]
        """

        for i in range(len(self.models)):
            if model == self.models[i].__class__:
                return self.models[i]
        raise ValueError("ModelDB does not contain an instance of the model {}.".format(model.__name__))

    def remove_model(self, model):
        """
        Remove a model from ModelDB.

        Parameters
        ----------
        model:
            Model class (*not* an instance) to remove. Example usage:

            .. code-block:: python

                from trait2d.analysis.models import ModelConfined
                ModelDB().remove_model(ModelConfined)
        """
        for i in range(len(self.models)):
            if model == self.models[i].__class__:
                self.models.pop(i)
                return
        raise ValueError("ModelDB does not contain an instance of the model {}.".format(model.__name__))

    def cleanup(self):
        """
        Remove all models from ModelDB. It is good practice to call this
        at the end of your scripts since other scripts might share the same
        instance.
        """
        self.models = []

class ListOfTracks:
    """Create an object that can hold multiple tracks and analyze them in bulk.

    Parameters
    ----------
    tracks : list
        A Python list containing the tracks as `Track` objects.
    """

    def __init__(self, tracks: list = None):
        self._tracks = tracks

    @classmethod
    def from_file(cls, filename, format=None, col_name_x='x', col_name_y='y', col_name_t='t', col_name_id='id', unit_length='metres', unit_time='seconds'):
        """Create a ListOfTracks from a file containing multiple tracks. Currently only supports '.csv' files.
        The file must contain the fields 'x', 'y', 't' as well as 'id'. Different column names can also be
        specified using the appropriate arguments.

        Parameters
        ----------
        filename: str
            Name of the file.
        format: str
            Either 'csv' or 'json' or 'pcl'. Only csv is implemented at the moment.
        col_name_x: str
            Column title of x positions.
        col_name_y: str
            Column title of y positions.
        col_name_t: str
            Column title of time.
        col_name_id: str
            Column title of track IDs.
        unit_length: str
            Length unit of track data. Either 'metres', 'millimetres', 'micrometres' or 'nanometres'.
        unit_time: str
            Time unit of track data. Either 'seconds', 'milliseconds', 'microseconds' or 'nanoseconds'.
        id: int
            Track ID in case the file contains more than one track.

        Raises
        ------
        LoadTrackIdNotFoundError
            When no track with the given id is found
        LodTrackIdMissingError
            When the file contains multiple tracks but no id is specified.
        """
        df = pd.read_csv(filename)
        ids = df[col_name_id].unique()
        tracks = []
        for id in ids:
            tracks.append(Track.from_dataframe(df, col_name_x, col_name_y, col_name_t, col_name_id, unit_length, unit_time, id))
        return cls(tracks)

    def __repr__(self):
        return ("<%s instance at %s>\n"
                "Number of tracks: %s\n") % (self.__class__.__name__,
                                             id(self),
                                             len(self._tracks))

    def get_tracks(self):
        """Return a Python list that contains the tracks.

        Returns
        -------
        tracks : list
            List of tracks. Each element will be a `Track` object.
        """
        return self._tracks

    def get_track(self, idx):
        """Return a single track at the specified index.

        Parameters
        ----------
        idx : int
            Index of track.

        Returns
        -------
        track: Track
            Track at index idx.
        """
        return self._tracks[idx]

    def get_track_by_id(self, id):
        """Return a single track with the specified ID.
        If there are multiple tracks with the same ID,
        only the first track is returned.

        Parameters
        ----------
        id: int
            ID of the track.

        Returns
        -------
        track: Track
            Track with the specified ID.
        """
        for track in self._tracks:
            if track.get_id() == id:
                return track

    def get_sublist(self, model):
        """Return a new ListOfTracks containing only tracks categorized as
        the specified model using ADC analysis.

        Parameters
        ----------
        model
            Class (*not* an instance) of the model the tracks should be categorized as.
            Predefined models can be found at `trait2d.analysis.models`.

        Returns
        -------
        track_list: ListOfTracks
            ListOfTracks containing the tracks that meet the criteria.
            Note: The tracks in the new list will still be references to to
            the tracks in the original list.
        """
        track_list = []
        for track in self._tracks:
            if track.get_adc_analysis_results() is not None:
                if track.get_adc_analysis_results()["best_model"] == model.__name__:
                    track_list.append(track)
        return ListOfTracks(track_list)

    def plot_trajectories(self, cmap="plasma"):
        """Plot all trajectories.

        Parameters
        ----------
        cmap : str
            Name of the colormap to use (see https://matplotlib.org/tutorials/colors/colormaps.html
            for a list of possible values)
        """
        import matplotlib.pyplot as plt
        from matplotlib.collections import LineCollection
        from matplotlib.ticker import FuncFormatter
        from matplotlib.cm import get_cmap
        cmap = get_cmap(cmap)
        plt.figure()
        ax = plt.gca()
        xmin = 0
        xmax = 0
        ymin = 0
        ymax = 0
        for track in self._tracks:
            segs = []
            colors = []
            x = track.get_x()
            y = track.get_y()
            t = track.get_t()
            x -= x[0]
            y -= y[0]
            t -= t[0]
            tmax = t.max()
            tmin = t.min()
            tdif = tmax - tmin
            for i in range(1, t.size):
                segs.append([(x[i-1], y[i-1]),
                             (x[i], y[i])])
                colors.append(cmap((t[i] - tmin) / tdif))
            lc = LineCollection(segs, colors=colors)
            ax.add_collection(lc)
            xmin = min(xmin, x.min())
            xmax = max(xmax, x.max())
            ymin = min(ymin, y.min())
            ymax = max(ymax, y.max())
        ax.set_title("Plot of {} trajectories".format(len(self._tracks)))
        xspan = xmax - xmin
        yspan = ymax - ymin
        ax.set_xlim(xmin - 0.1 * xspan, xmax + 0.1 * xspan)
        ax.set_ylim(ymin - 0.1 * yspan, ymax + 0.1 * yspan)
        ax.axhline(0, linewidth=0.5, color='black', zorder=-1)
        ax.axvline(0, linewidth=0.5, color='black', zorder=-1)
        ax.set_aspect('equal', 'datalim')
        ax.xaxis.set_major_formatter(
            FuncFormatter(lambda x, pos: "%d" % int(x * 1e9)))
        ax.yaxis.set_major_formatter(
            FuncFormatter(lambda x, pos: "%d" % int(x * 1e9)))
        ax.set_xlabel("x (nm)")
        ax.set_ylabel("y (nm)")
        plt.show()

    def normalize(self, **kwargs):
        """Normalize all tracks. All Track instances of the ListOfTracks will be
        replaced by NormalizedTrack instances containing individual information about
        the normalization.

        Parameters
        ----------
            Keyword arguments to be used by Track.normalized() for each track.
        """
        self._tracks = [track.normalized(**kwargs) for track in self._tracks]

    def msd_analysis(self, **kwargs):
        """Analyze all tracks using MSD analysis.

        Parameters
        ----------
        **kwargs
            Keyword arguments to be used by msd_analysis for each track.

        Returns
        -------
        list_failed
            List containing the indices of the tracks for which the analysis failed.
        """
        list_failed = []
        i = 0
        for track in self._tracks:
            try:
                track.msd_analysis(**kwargs)
            except:
                list_failed.append(i)
            i += 1

        if len(list_failed) > 0:
            warnings.warn("MSD analysis failed for {}/{} tracks. \
                Consider raising the maximum function evaluations using \
                the maxfev keyword argument. \
                To get a more detailed stacktrace, run the MSD analysis \
                for a single track.".format(len(list_failed), len(self._tracks)))

        return list_failed

    def adc_analysis(self, **kwargs):
        """Analyze all tracks using ADC analysis.

        Parameters
        ----------
        **kwargs
            Keyword arguments to be used by adc_analysis for each track.

        Returns
        -------
        list_failed
            List containing the indices of the tracks for which the analysis failed.
        """
        list_failed = []
        i = 0
        for track in self._tracks:
            try:
                track.adc_analysis(**kwargs)
            except:
                list_failed.append(i)
            i+=1
        
        if len(list_failed) > 0:
            warnings.warn("ADC analysis failed for {}/{} tracks. "
                "Consider raising the maximum function evaluations using "
                "the maxfev keyword argument. "
                "To get a more detailed stacktrace, run the ADC analysis "
                "for a single track.".format(len(list_failed), len(self._tracks)))

        return list_failed

    def adc_summary(self, avg_only_params = False, interpolation = False, plot_msd = False, plot_dapp = False, plot_pie_chart = False):
        """Average tracks by model and optionally plot the results.

        Parameters
        ----------
        avg_only_params: bool
            Only average the model parameters but not D_app and MSD
        interpolation: bool
            User linear interpolation of averaging. Has to be used when
            not all tracks have the same uniform time step size.
        plot_msd: bool
            Plot the averaged MSD for each model.
        plot_dapp: bool
            Plot the averaged D_app for each model.
        plot_pie_chart: bool
            Plot a pie chart showing the relative fractions of each model.

        Returns
        -------
        results : dict
            Relative shares and averaged values of each model.
        """

        if avg_only_params and (plot_msd or plot_dapp):
            warnings.warn("avg_only_params is True. plot_msd or plot_dapp will have no effect.")

        track_length = 0
        max_t = 0.0
        t = None
        for track in self._tracks:
            if track.get_t()[-1] > max_t:
                max_t = track.get_t()[-1]
                track_length = track.get_x().size
                t = track.get_t()

        average_D_app = {}
        average_MSD = {}
        average_params = {}
        sampled = {}
        counter = {}

        dt = t[1] - t[0]

        k = 0
        for track in self._tracks:
            k += 1
            if track.get_t()[1] - track.get_t()[0] != dt and not avg_only_params and not interpolation:
                raise ValueError("Cannot average MSD and D_app: Encountered track with incorrect time step size! "
                                 "(Got {}, expected {} for track {}.) Use the flag avg_only_params = True or "
                                 "enable interpolation with interpolation = True.".format(
                    track.get_t()[1] - track.get_t()[0], dt, k + 1))

        # Parameter averaging
        for track in self._tracks:
            if track.get_adc_analysis_results() is None:
                continue
            model = track.get_adc_analysis_results()["best_model"]
            if model is not None:
                if not model in average_params.keys():
                    average_params[model] = len(track.get_adc_analysis_results()["fit_results"][model]["params"]) * [0.0]
                    counter[model] = 0
                counter[model] += 1
                average_params[model] += track.get_adc_analysis_results()["fit_results"][model]["params"]
        
        for model in average_params.keys():
            average_params[model] /= counter[model]
        
        k = 0
        for track in self._tracks:
            k += 1
            if track.get_t()[1] - track.get_t()[0] != dt and not avg_only_params and not interpolation:
                raise ValueError("Cannot average MSD and D_app: Encountered track with incorrect time step size! "
                                 "(Got {}, expected {} for track {}.) Use the flag avg_only_params = True or "
                                 "enable interpolation with interpolation = True.".format(
                    track.get_t()[1] - track.get_t()[0], dt, k + 1))

        if not avg_only_params:
            for track in self._tracks:
                if track.get_adc_analysis_results() is None:
                    continue

                model = track.get_adc_analysis_results()["best_model"]

                D_app = np.zeros(track_length - 3)
                MSD = np.zeros(track_length - 3)
                if interpolation:
                    interp_MSD = interpolate.interp1d(track.get_t()[0:-3], track.get_msd(), bounds_error = False, fill_value = 0)
                    interp_D_app = interpolate.interp1d(track.get_t()[0:-3], track.get_adc_analysis_results()["Dapp"], bounds_error = False, fill_value = 0)
                    MSD = interp_MSD(t[0:-3])
                    D_app = interp_D_app(t[0:-3])
                else:
                    D_app[0:track.get_adc_analysis_results()["Dapp"].size] = track.get_adc_analysis_results()["Dapp"]
                    MSD[0:track.get_msd().size] = track.get_msd()
                mask = np.zeros(track_length - 3)
                np.put(mask, np.where(MSD != 0.0), 1)


                if not model in average_D_app.keys():
                    average_D_app[model] = np.zeros(track_length - 3)
                    average_MSD[model] = np.zeros(track_length - 3)
                    sampled[model] = np.zeros(track_length - 3)

                average_D_app[model] += D_app
                average_MSD[model] += MSD
                sampled[model] += mask

        counter_sum = 0
        for model in counter:
            counter_sum += counter[model]
            if counter[model]:
                average_D_app[model] /= sampled[model]
                average_MSD[model] /= sampled[model]

        if counter_sum == 0:
            warnings.warn("No tracks are categorized!")

        sector = {}
        for model in counter:
            sector[model] = counter[model] / len(self._tracks)
        sector["not catergorized"] = (len(self._tracks) - counter_sum) / len(self._tracks)

        if plot_msd and not avg_only_params:
            import matplotlib.pyplot as plt
            plt.figure()
            ax = plt.gca()
            ax.set_xlabel("t")
            ax.set_ylabel("Average MSD")
            for model in counter:
                ax.semilogx(t[0:-3], average_MSD[model], label=model)
            ax.legend()

        if plot_dapp and not avg_only_params:
            import matplotlib.pyplot as plt
            plt.figure()
            min_val = 9999999.9
            max_val = 0.0
            ax = plt.gca()
            ax.set_xlabel("t")
            ax.set_ylabel("Average D_app")
            for model in counter:
                new_min = np.min(average_D_app[model])
                new_max = np.max(average_D_app[model])
                min_val = min(min_val, new_min)
                max_val = max(max_val, new_max)
                l, = ax.semilogx(t[0:-3], average_D_app[model], label=model)
                r = average_params[model]
                for c in ModelDB().models:
                    if c.__class__.__name__ == model:
                        m = c
                pred = m(t, *r)
                plt.semilogx(t[0:-3], pred[0:-3], linestyle='dashed', color=l.get_color())
            ax.set_ylim(0.95*min_val, 1.05*max_val)
            ax.legend()

        if plot_pie_chart:
            import matplotlib.pyplot as plt
            plt.figure()
            ax = plt.gca()
            ax.pie(sector.values(),
                   labels=sector.keys())

        return {"sectors": sector,
                "average_params": average_params, 
                "average_msd": average_MSD,
                "average_dapp": average_D_app}

    def get_msd(self, interpolation = False): # Get average MSD
        """Get the mean squared displacement averaged over all tracks.

        Parameters
        ----------
        interpolation: bool
            Linearly interpolate all msd values over the time points of the first track.
            Use when working with differently spaced tracks.

        Returns
        -------
        t : ndarray
            1-dimensional array containing time points at which msd was sampled
        msd: ndarray
            1-dimensional array containing averaged msd values
        """
        track_length = 0
        max_t = 0.0
        t = None
        for track in self._tracks:
            if track.get_t()[-1] > max_t:
                max_t = track.get_t()[-1]
                track_length = track.get_x().size
                t = track.get_t()
        average_MSD = np.zeros(track_length - 3)
        average_MSD_err = np.zeros(track_length - 3)
        sampled = np.zeros(track_length - 3)
        for track in self._tracks:
            if track.get_msd() is None:
                continue

            MSD = np.zeros(track_length - 3)
            MSD_err = np.zeros(track_length - 3)
            if interpolation:
                interp_MSD = interpolate.interp1d(track.get_t()[0:-3], track.get_msd(), bounds_error = False, fill_value = 0)
                interp_MSD_err = interpolate.interp1d(track.get_t()[0:-3], track.get_msd(), bounds_error = False, fill_value = 0)
                MSD = interp_MSD(t[0:-3])
                MSD_err = interp_MSD_err(t[0:-3])
            else:
                MSD[0:track.get_msd().size] = track.get_msd()
                MSD_err[0:track.get_msd().size] = track._msd_error
            mask = np.zeros(track_length - 3)
            np.put(mask, np.where(MSD != 0.0), 1)

            average_MSD += MSD
            average_MSD_err += MSD_err
            sampled += mask
        average_MSD /= sampled
        average_MSD_err /= sampled
        return t[0:-3], average_MSD, average_MSD_err

    def plot_msd(self, interpolation = False):
        """Plot the mean squared displacement averaged over all tracks.

        Parameters
        ----------
        interpolation: bool
            Linearly interpolate all msd values over the time points of the first track.
            Use when working with differently spaced tracks.
        """

        t, msd, err= self.get_msd(interpolation)
        import matplotlib.pyplot as plt
        plt.figure()
        plt.grid(linestyle='dashed', color='grey')
        plt.xlabel("t")
        plt.ylabel("Average MSD")
        plt.semilogx(t, msd, color='black')
        plt.fill_between(t, msd-err, msd+err, color='black', alpha=0.5)

    def get_dapp(self, interpolation = False):
        """Get the apparent diffusion coefficient averaged over all tracks.

        Parameters
        ----------
        interpolation: bool
            Linearly interpolate all D_app values over the time points of the first track.
            Use when working with differently spaced tracks.

        Returns
        -------
        t : ndarray
            1-dimensional array containing time points at which msd was sampled
        average_dapp: ndarray
            1-dimensional array containing averaged D_app values
        """
        track_length = 0
        max_t = 0.0
        t = None
        for track in self._tracks:
            if track.get_t()[-1] > max_t:
                max_t = track.get_t()[-1]
                track_length = track.get_x().size
                t = track.get_t()
        average_dapp = np.zeros(track_length - 3)
        sampled = np.zeros(track_length - 3)
        for track in self._tracks:
            if track.get_adc_analysis_results() is None:
                continue

            D_app = np.zeros(track_length - 3)
            if interpolation:
                interp_D_app = interpolate.interp1d(track.get_t()[0:-3], track.get_adc_analysis_results()["Dapp"], bounds_error = False, fill_value = 0)
                D_app = interp_D_app(t[0:-3])
            else:
                D_app[0:track.get_adc_analysis_results()["Dapp"].size] = track.get_adc_analysis_results()["Dapp"]
            mask = np.zeros(track_length - 3)
            np.put(mask, np.where(D_app != 0.0), 1)

            average_dapp += D_app
            sampled += mask
        average_dapp /= sampled
        return t[0:-3], average_dapp

    def plot_dapp(self):
        """Plot the apparent diffusion coefficient averaged over all tracks.

        Parameters
        ----------
        interpolation: bool
            Linearly interpolate all D_app values over the time points of the first track.
            Use when working with differently spaced tracks.
        """
        t, dapp = self.get_dapp()
        import matplotlib.pyplot as plt
        plt.figure()
        ax = plt.gca()
        ax.set_xlabel("t")
        ax.set_ylabel("Average Dapp")
        ax.semilogx(t, dapp)
        ax.legend()

    def average(self, interpolation=False):
        """Get the mean squared displacement averaged over all tracks.

        Parameters
        ----------
        interpolation: bool
            Linearly interpolate all msd values over the time points of the first track.
            Use when working with differently spaced tracks.
        use_averaged_msd: bool
            Initialize the track with a MSD curve which is the average of all MSD curves
            in the list. This is *not* neccessarily equal to the result of performing
            `calculate_msd` on the averaged track.

        Returns
        -------
        track : Track
            A single track containing the average 
        """
        t, msd, msd_err = self.get_msd()
        return MSDTrack(msd, msd_err, t)

class Track:
    """Create a track that can hold trajectory and analysis information.

    Parameters
    ----------
    x: array_like
        x coordinates of trajectory.
    y: array_like
        y coordinates of trajectory.
    t: array_like
        time coordinates of trajectory.
    """

    def __init__(self, x=None, y=None, t=None, id=None):
        self._x = np.array(x, dtype=float)
        self._y = np.array(y, dtype=float)
        self._t = np.array(t, dtype=float)

        self._id = id

        self._msd = None
        self._msd_error = None

        self._msd_analysis_results = None
        self._adc_analysis_results = None

    @classmethod
    def from_dict(cls, dict):
        """Create a track from a dictionary.
        Parameters
        ----------
        dict: dict
            Dictionary of the track. Has to contain the fields "x", "y" and "t".
        """
        return cls(dict["x"], dict["y"], dict["t"])

    @classmethod
    def from_dataframe(cls, df, col_name_x='x', col_name_y='y', col_name_t='t', col_name_id='id', unit_length='metres', unit_time='seconds', id=None):
        """Create a single track from a DataFrame. Currently only supports '.csv' tracks.
        The DataFrame must contain the fields 'x', 'y', 't' as well as 'id'. Different column names can also be
        specified using the appropriate arguments.

        Parameters
        ----------
        df: pandas.DataFrame
            DataFrame containing the data.
        col_name_x: str
            Column title of x positions.
        col_name_y: str
            Column title of y positions.
        col_name_t: str
            Column title of time.
        col_name_id: str
            Column title of track IDs.
        unit_length: str
            Length unit of track data. Either 'metres', 'millimetres', 'micrometres' or 'nanometres'.
        unit_time: str
            Time unit of track data. Either 'seconds', 'milliseconds', 'microseconds' or 'nanoseconds'.
        id: int
            Track ID in case the file contains more than one track.

        Raises
        ------
        LoadTrackIdNotFoundError
            When no track with the given id is found
        LodTrackIdMissingError
            When the file contains multiple tracks but no id is specified.
        """
        length_factor = None
        if unit_length == "metres":
            length_factor = 1
        elif unit_length == "millimetres":
            length_factor = 1e-3
        elif unit_length == "micrometres":
            length_factor = 1e-6
        elif unit_length == "nanometres":
            length_factor = 1e-9
        else:
            raise ValueError(
                "unit must be metres, millimetres, micrometres, or nanometres")

        time_factor = None
        if unit_time == "seconds":
            time_factor = 1
        elif unit_time == "milliseconds":
            time_factor = 1e-3
        elif unit_time == "microseconds":
            time_factor = 1e-6
        elif unit_time == "nanoseconds":
            time_factor = 1e-9
        else:
            raise ValueError(
                "unit must be metres, millimetres, micrometres, or nanometres")

        x = None
        y = None
        t = None
        if col_name_id in df:
            if np.min(df[col_name_id]) == np.max(df[col_name_id]):
                id = np.min(df[col_name_id])  # If there is only one id, we just select it
            if id == None:
                raise LoadTrackMissingIdError("The file seems to contain more than one track. Please specify a track id using the keyword argument id.")
            df = df.loc[df[col_name_id] == int(id)]
            if df.empty:
                raise LoadTrackIdNotFoundError("There is no track associated with the specified id!")

        x = np.array(df[col_name_x])
        y = np.array(df[col_name_y])
        t = np.array(df[col_name_t])

        return cls(x * length_factor, y * length_factor, t * time_factor, id=id)

    @classmethod
    def from_file(cls, filename, format=None, col_name_x='x', col_name_y='y', col_name_t='t', col_name_id='id', unit_length='metres', unit_time='seconds', id=None):
        """Create a single track from a file. Currently only supports '.csv' tracks.
        The DataFrame must contain the fields 'x', 'y', 't' as well as 'id'. Different column names can also be
        specified using the appropriate arguments.

        Parameters
        ----------
        filename: str
            Name of the file.
        format: str
            Either 'csv' or 'json' or 'pcl'. Only csv is implemented at the moment.
        col_name_x: str
            Column title of x positions.
        col_name_y: str
            Column title of y positions.
        col_name_t: str
            Column title of time.
        col_name_id: str
            Column title of track IDs.
        unit_length: str
            Length unit of track data. Either 'metres', 'millimetres', 'micrometres' or 'nanometres'.
        unit_time: str
            Time unit of track data. Either 'seconds', 'milliseconds', 'microseconds' or 'nanoseconds'.
        id: int
            Track ID in case the file contains more than one track.

        Raises
        ------
        LoadTrackIdNotFoundError
            When no track with the given id is found
        LodTrackIdMissingError
            When the file contains multiple tracks but no id is specified.
        """
        if format == None:
            format = os.path.splitext(filename)[1].replace(".", "")
            if format == "":
                raise ValueError(
                    'You must supply a format "csv", "json" or "pcl" if the file has no extension.')

        if format != "csv" and format != "json" and format != "pcl":
            raise ValueError("Unknown format: {}".format(format))
        if format == "csv":
            df = pd.read_csv(filename)
            return cls.from_dataframe(df, col_name_x, col_name_y, col_name_t, col_name_id, unit_length, unit_time, id)
        elif format == "json":
            # TODO: .json-specific import
            raise NotImplementedError(
                "Import from json is not yet implemented.")
        elif format == "pcl":
            # TODO: .pcl-specific import
            raise NotImplementedError(
                "Import from pcl is not yet implemented.")

    def __repr__(self):
        return ("<%s instance at %s>\n"
                "------------------------\n"
                "Track length:%s\n"
                "Track ID:%s\n"
                "------------------------\n"
                "MSD calculated:%s\n"
                "MSD analysis done:%s\n"
                "ADC analysis done:%s\n") % (
            self.__class__.__name__,
            id(self),
            str(self._t.size).rjust(11, ' '),
            str(self._id).rjust(15, ' '),
            str(self.is_msd_calculated()).rjust(9, ' '),
            str(self._msd_analysis_results is not None).rjust(6, ' '),
            str(self._adc_analysis_results is not None).rjust(6, ' ')
        )

    from ._msd import msd_analysis, get_msd_analysis_results,\
                      delete_msd_analysis_results, plot_msd_analysis_results
    from ._adc import adc_analysis, get_adc_analysis_results,\
                      delete_adc_analysis_results, plot_adc_analysis_results

    def plot_trajectory(self, cmap='plasma'):
        """Plot the trajectory.

        Parameters
        ----------
        cmap : str
            Name of the colormap to use (see https://matplotlib.org/tutorials/colors/colormaps.html
            for a list of possible values)
        """
        import matplotlib.pyplot as plt
        from matplotlib.collections import LineCollection
        from matplotlib.ticker import FuncFormatter
        from matplotlib.cm import get_cmap
        cmap = get_cmap(cmap)
        plt.figure()
        ax = plt.gca()
        segs = []
        colors = []

        tmax = self._t.max()
        tmin = self._t.min()
        tdif = tmax - tmin

        for i in range(1, self._t.size):
            segs.append([(self._x[i-1], self._y[i-1]),
                         (self._x[i], self._y[i])])
            colors.append(
                cmap((self._t[i] - tmin) / tdif))
        lc = LineCollection(segs, colors=colors)
        ax.axhline(self._y[0], linewidth=0.5, color='black', zorder=-1)
        ax.axvline(self._x[0], linewidth=0.5, color='black', zorder=-1)
        ax.add_collection(lc)
        ax.set_title("Trajectory")
        xspan = self._x.max() - self._x.min()
        yspan = self._y.max() - self._y.min()
        ax.set_xlim(self._x.min() - 0.1 * xspan, self._x.max() + 0.1 * xspan)
        ax.set_ylim(self._y.min() - 0.1 * yspan, self._y.max() + 0.1 * yspan)
        ax.set_aspect('equal', 'datalim')
        ax.xaxis.set_major_formatter(
            FuncFormatter(lambda x, pos: "%d" % int(x * 1e9)))
        ax.yaxis.set_major_formatter(
            FuncFormatter(lambda x, pos: "%d" % int(x * 1e9)))
        ax.set_xlabel("x (nm)")
        ax.set_ylabel("y (nm)")
        plt.show()

    def plot_msd(self):
        t = self._t[0:-3]
        msd = self._msd
        err = self._msd_error
        import matplotlib.pyplot as plt
        plt.figure()
        plt.grid(linestyle='dashed', color='grey')
        plt.xlabel("t")
        plt.ylabel("MSD")
        plt.semilogx(t, msd, color='black')
        plt.fill_between(t, msd-err, msd+err, color='black', alpha=0.5)


    def get_x(self):
        """Return x coordinates of trajectory."""
        return self._x

    def get_y(self):
        """Return y coordinates of trajectory."""
        return self._y

    def get_t(self):
        """Return time coordinates of trajectory."""
        return self._t

    def get_id(self):
        """Return ID of the track."""
        return self._id

    def get_size(self):
        """Return number of points of the trajectory."""
        return self._t.size

    def get_trajectory(self):
        """Returns the trajectory as a dictionary."""
        return {"t": self._t.tolist(), "x": self._x.tolist(), "y": self._y.tolist()}

    def is_msd_calculated(self):
        """Returns True if the MSD of this track has already been calculated."""
        return self._msd is not None

    def get_msd(self):
        """Returns the MSD values of the track."""
        return self._msd

    def get_msd_error(self):
        """Returns the MSD error values of the track."""
        return self._msd_error

    def normalized(self, normalize_t = True, normalize_xy = True):
        """Normalize the track.

        Parameters
        ----------
        normalize_t: bool
            Normalize the time component of the track by setting the first time in the
            track to zero.
        normalize_xy: bool
            Normalize the x and y coordinates of the track by setting the initial
            position in the track to zero.

        Returns
        -------
        Instance of NormalizedTrack containing the normalized track data.
        """

        x = self._x
        y = self._y
        t = self._t
        
        xmin = 0.0
        ymin = 0.0
        tmin = 0.0

        # Normalizing the coordinates
        if normalize_xy:
            xmin = x.min()
            ymin = y.min()
            x = x - xmin
            y = y - ymin
        if normalize_t:
            tmin = t.min()
            t = t - tmin

        # Create normalized Track object
        return NormalizedTrack(x, y, t, xmin, ymin, tmin, id=self._id)

    def calculate_msd(self):
        """Calculates the mean squared displacement of the track. Result is stored in the Track object.
        """

        def circular_matrix(ar):
            a = np.concatenate(( ar, ar[:-1] ))
            L = len(ar)
            n = a.strides[0]
            return np.lib.stride_tricks.as_strided(a[L-1:], (L,L), (-n,n))

        xc = circular_matrix(self._x)
        yc = circular_matrix(self._y)

        sdm = (self._x - xc)**2 + (self._y - yc)**2
        sdm = np.triu(sdm, 1)[:-3]

        # MSD is the mean of each upper triangular row (except the last)
        d = np.arange(sdm.shape[1] - 1, 2, -1)
        
        self._msd = np.sum(sdm, axis=1) / d
        self._msd_error = np.sqrt(np.sum(np.triu((sdm - self._msd[:, None])**2, 1), axis=1) / d**2) # stddev weighted by inverse square of msd length

    def _categorize(self, Dapp, J, Dapp_err = None, R: float = 1/6, fraction_fit_points: float = 0.25, fit_max_time: float=None, maxfev=1000, enable_log_sampling = False, log_sampling_dist = 0.2, weighting = 'error'):
        if fraction_fit_points > 0.25:
            warnings.warn(
                "Using too many points for the fit means including points which have higher measurment errors.")

        dt = self._t[1] - self._t[0]
        T = J * dt

        # Get number of points for fit from either fit_max_time or fraction_fit_points
        if fit_max_time is not None:
            n_points = int(np.argwhere(T < fit_max_time)[-1])
        else:
            n_points = np.argmax(J > fraction_fit_points * J[-1])

        cur_dist = 0
        idxs = []
        if enable_log_sampling:
            # Get indexes that are (approximately) logarithmically spaced
            idxs.append(0)
            for i in range(1, n_points):
                cur_dist += np.log10(T[i]/T[i-1])
                if cur_dist >= log_sampling_dist:
                    idxs.append(i)
                    cur_dist = 0
        else:
            # Get every index up to n_points
            idxs = np.arange(0, n_points, dtype=int)

        error = None
        if not Dapp_err is None:
            error = Dapp_err[idxs]

        # Perform fits for all included models
        fit_results = {}

        bic_min = 999.9
        category = None
        sigma = None
        if weighting == 'error':
            sigma = error
        elif weighting == 'inverse_variance':
            sigma = np.power(error, 2.0)
        elif weighting == 'variance':
            sigma = 1 / np.power(error, 2.0)
        elif weighting == 'disabled':
            sigma = None
        else:
            raise ValueError("Unknown weighting method: {}. Possible values are: 'error', 'variance', 'inverse_variance', and 'disabled'.".format(weighting))

        for model in ModelDB().models:
            model.R = R
            model.dt = dt
            model_name = model.__class__.__name__

            r = optimize.curve_fit(model, T[idxs], Dapp[idxs], p0 = model.initial,
                        sigma = sigma, maxfev = maxfev, method='trf', bounds=(model.lower, model.upper))

            perr = np.sqrt(np.diag(r[1]))
            pred = model(T, *r[0])
            bic = BIC(pred[idxs], Dapp[idxs], len(r[0]), len(idxs))
            if bic < bic_min:
                bic_min = bic
                category = model_name

            from scipy.stats import kstest
            test_results = kstest(Dapp[idxs], pred[idxs], N = len(idxs))

            fit_results[model_name] = {"params": r[0], "errors": perr, "bic" : bic, "KSTestStat": test_results[0], "KStestPValue": test_results[1]}

        # Calculate the relative likelihood for each model
        for model in ModelDB().models:
            model_name = model.__class__.__name__
            rel_likelihood = np.exp((-fit_results[model_name]["bic"] + bic_min) * 0.5)
            fit_results[model_name]["rel_likelihood"] = rel_likelihood

        fit_indices = idxs
        return category, fit_indices, fit_results


class NormalizedTrack(Track):
    """A track with normalized coordinates and additional information about the normalization."""

    def __init__(self, x=None, y=None, t=None, xmin=None, ymin=None, tmin=None, id=None):
        Track.__init__(self, x, y, t, id=id)
        self._xmin = xmin
        self._ymin = ymin
        self._tmin = tmin

class MSDTrack(Track):
    """A special track class that holds an MSD curve but no trajectory data. Can be used for analysis like an ordinary :class:`trait2d.analysis.Track` object but certain functions might not be available."""

    def __init__(self, msd, msd_err, t, id=None):
        t = np.append(t, np.zeros(3))
        Track.__init__(self, t=t, id=id)
        self._msd = msd
        self._msd_error = msd_err

    def normalized(normalize_t=True, normalize_xy=True):
        TypeError("A MSDTrack instance cannot be normalized.")

    def calculate_msd(self):
        TypeError("It is not possible to calculate the MSD of a MSDTrack instance.")

    def plot_trajectory(self):
        TypeError("It is not possible to plot the trajectory of a MSDTrack instance.")

    def __repr__(self):
        return ("<%s instance at %s>\n"
                "------------------------\n"
                "Track length:%s\n"
                "Track ID:%s\n"
                "------------------------\n"
                "MSD analysis done:%s\n"
                "ADC analysis done:%s\n") % (
            self.__class__.__name__,
            id(self),
            str(self._t.size).rjust(11, ' '),
            str(self._id).rjust(15, ' '),
            str(self._msd_analysis_results is not None).rjust(6, ' '),
            str(self._adc_analysis_results is not None).rjust(6, ' ')
        )


def BIC(pred: list, target: list, k: int, n: int):
    """Bayesian Information Criterion
    Parameters
    ----------
    pred: list
        Model prediction
    target: list
        Model targe

    k: int
        Number of free parameters in the fit
    n: int
        Number of data points used to fit the model
    Returns
    -------
    bic : float
        Bayesian Information Criterion
    """
    # Compute RSS
    RSS = np.sum((np.array(pred) - np.array(target)) ** 2)
    bic = k * np.log(n) + n * np.log(RSS / n)
    return bic


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
TRAIT tracker GUI 

"""

import matplotlib
matplotlib.use('TkAgg') # This is a fix to use the GUI on Mac

import sys
from trait2d.detectors import Detectors
from trait2d.movie_processor import background_substraction
from trait2d.tracker import Tracker

import skimage
from skimage import io
import matplotlib.pyplot as plt

import numpy as np
import cv2
import tkinter as tk
from tkinter import filedialog
import csv

# for plotting
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

# default Matplotlib key bindings
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure import Figure
import imageio

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning) 

class MainVisual(tk.Frame):
    '''
    GUI for TRAIT tracker
    '''

    def __init__(self, master):

        #define the window
        tk.Frame.__init__(self, master)
        self.master = master
        master.title("TRAIT tracker")
        master.configure(background='white')
        
        # get the monitor size
        self.dpi=100
        self.monitor_width=master.winfo_screenwidth()
        self.monitor_height = master.winfo_screenheight()
        self.button_size=int(self.monitor_width/25)
        self.figsize_value=(int(self.monitor_height/3/self.dpi), int(self.monitor_height/3/self.dpi)) # parameters for the figure
        self.scale_length=int(self.monitor_height/3)
        
        master.protocol('WM_DELETE_WINDOW', self.close_app)

        # # # parameters # # #
        #general
        self.movie_file=" " # path to the movie file
        self.movie=np.ones((1,400,400))*0.8 # matrix with data
        self.movie_processed=np.ones((1, 400,400))*0.8 # matrix with processed data
        self.movie_length=0 # length of the original movie
        self.frame_pos=0 # frame location
        
        
        self.img_resolution=1 # nm/pix
        self.frame_rate=1 # sec/frame

        
        
        # detection and tracking parameters
        self.maximum_diameter=10 # size of the field for gaussian fitting
        self.sigma=6. # defines gaussian in spot-enhancing filter
        self.threshold=4. # defines threshold in spot-enhancing filter
        self.min_peak=0.2 # defines parameter for local maxima [0,1]

        self.max_dist=15 # number of pixels between two possible connections
        self.frame_gap=15 # maximum number of possible consequently skipped frames in a track
        self.spot_switch=0 # spot type: 0 - dark spot, 1 - light spot
        
        self.min_track_length=100 # track with length less than the value will be removed

        # list of colors for trajectory plots
        self.color_list=[(200, 0, 0), (200, 0, 127), (0, 0, 255), (200, 155, 0),
                    (0, 255, 59), (255, 10, 120), (255, 127, 255),
                    (127, 0, 255), (0, 255, 0), (177, 0, 20),
                    (12, 200, 0), (0, 114, 255), (255, 20, 0),
                    (0, 255, 255), (255, 100, 100), (255, 127, 255),
                    (127, 0, 255), (127, 0, 127)]

        self.tracks_data=[] # data of the tracks prepared for the csv file
        
     # # # # # # menu to choose files and set tracker parameters # # # # # #

        # button to select movie
        self.button1 = tk.Button(text="       Select movie file       ", command=self.select_movie, width=int(self.button_size/3), bg='gray')
        self.button1.grid(row=0, column=1, columnspan=3, pady=5)

        # show selected movie name
        self.lbl1 = tk.Label(master=root, text="Selected file: "+self.movie_file, bg='white')
        self.lbl1.grid(row=1, column=1, columnspan=3, pady=5)

        # button for preprocessing step
        self.button2 = tk.Button(text="    Run pre-processing step    ", command=self.processing, width=int(self.button_size/3), bg='gray')
        self.button2.grid(row=2, column=1, columnspan=3, pady=5)

        # setting tracker parameters

        lbl1 = tk.Label(master=root, text="PARAMETERS: ", width=int(self.button_size/2), bg='white')
        lbl1.grid(row=3, column=1, columnspan=3, pady=5)


        lbl3 = tk.Label(master=root, text="SEF: sigma", width=int(self.button_size/2), bg='white')
        lbl3.grid(row=4, column=1)
        v = tk.StringVar(root, value=str(self.sigma))
        self.param2_sigma = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param2_sigma.grid(row=4, column=2)

        lbl4 = tk.Label(master=root, text="SEF: threshold [0.01,10]", width=int(self.button_size/2), bg='white')
        lbl4.grid(row=5, column=1)
        v = tk.StringVar(root, value=str(self.threshold))
        self.param3_threshold = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param3_threshold.grid(row=5, column=2)

        lbl5 = tk.Label(master=root, text="SEF: min peak value [0,1]", width=int(self.button_size/2), bg='white')
        lbl5.grid(row=6, column=1)
        v = tk.StringVar(root, value=str(self.min_peak))
        self.param4_peak = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param4_peak.grid(row=6, column=2)

        lbl2 = tk.Label(master=root, text="  Patch size (even number), px", width=int(self.button_size/2), bg='white')
        lbl2.grid(row=7, column=1)
        v = tk.StringVar(root, value=str(self.maximum_diameter))
        self.param1_diameter = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param1_diameter.grid(row=7, column=2)


        lbl6 = tk.Label(master=root, text="Linking: max distance, px", width=int(self.button_size/2), bg='white')
        lbl6.grid(row=8, column=1)
        v = tk.StringVar(root, value=str(self.max_dist))
        self.param5_distance = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param5_distance.grid(row=8, column=2)

        lbl6 = tk.Label(master=root, text="Linking: frame gap, frame", width=int(self.button_size/2), bg='white')
        lbl6.grid(row=9, column=1)
        v = tk.StringVar(root, value=str(self.frame_gap))
        self.param6_framegap = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param6_framegap.grid(row=9, column=2)

        lbl7 = tk.Label(master=root, text="Minimum track length, frames", width=int(self.button_size/2), bg='white')
        lbl7.grid(row=10, column=1)
        v = tk.StringVar(root, value=str(self.min_track_length))
        self.param7_framegap = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param7_framegap.grid(row=10, column=2)
        
        
        # resolution in time and space                
        res_lb = tk.Label(master=root, text=" Resolution (\u03BCm per pix) : ", width=int(self.button_size/2), bg='white')
        res_lb.grid(row=11, column=1)
        v = tk.StringVar(root, value=str(self.img_resolution))
        self.res_parameter = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.res_parameter.grid(row=11, column=2)
            
        lbl3 = tk.Label(master=root, text="Frame rate (sec per frame) : ", width=int(self.button_size/2), bg='white')
        lbl3.grid(row=12, column=1)
        v = tk.StringVar(root, value=str(self.frame_rate))
        self.frame_parameter = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.frame_parameter.grid(row=12, column=2)  



        # type of spots (dark or light)
        var = tk.IntVar() # the switch variable

        # variable update
        def update_monitor_switch():
            self.spot_switch=var.get()

        # spot type switch: # 0 - dark spot, 1 - light spot
        self.R1 = tk.Radiobutton(root, text=" Dark spot ", variable=var, value=0, bg='white', command =update_monitor_switch)
        self.R1.grid(row=13, column=1, pady=5)

        self.R2 = tk.Radiobutton(root, text=" Light spot ", variable=var, value=1, bg='white',command = update_monitor_switch ) #  command=sel)
        self.R2.grid(row=13, column=2, pady=5)
        

        #preview button
        self.button2 = tk.Button(text="    Preview    ", command=self.preview, width=int(self.button_size/3), bg='gray') #, height=30)
        self.button2.grid(row=14, column=1, columnspan=1,pady=5, padx=5)

        # button to run the tracker and save image sequence with plotted trakectories (for visualisation)
        self.button2 = tk.Button(text="    Run tracking    ", command=self.tracking, width=int(self.button_size/3), bg='gray')
        self.button2.grid(row=14, column=2, columnspan=1, pady=5, padx=5)

        # button to save csv file
        self.button2 = tk.Button(text="    Save data    ", command=self.save_data, width=int(self.button_size/3), bg='gray')
        self.button2.grid(row=15, column=2, columnspan=1, pady=5, padx=5)


        # show dark screen until movie is selected
        self.fig, self.ax = plt.subplots(1,1,figsize=self.figsize_value)
        plt.axis('off')
        self.fig.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
        
        self.show_frame()

        def show_values(v):
            self.frame_pos=int(v)
            self.show_frame() 
          
        self.scale_movie = tk.Scale(root, from_=0, to=self.movie_processed.shape[0]-1, tickinterval=int(self.movie_processed.shape[0]/5), length=self.scale_length, width=10, orient="horizontal", command=show_values)
        self.scale_movie.set(self.frame_pos)        
        self.scale_movie.grid(row=21, column=1, columnspan=3, pady=5, padx=5) 
        


    def show_frame(self , centers=[]):
        '''
        show current frame 
        '''
        
        self.ax.clear() # clean the plot 
        self.ax.imshow(self.movie_processed[self.frame_pos,:,:], cmap="gray")
        self.ax.axis('off')
        for point in centers:
            self.ax.plot(point[1], point[0],  "*r")
        
        # DrawingArea
        self.canvas = FigureCanvasTkAgg(self.fig, master=root)
        self.canvas.get_tk_widget().grid(row=20, column=1, columnspan=3,pady=5)
        self.canvas.draw()
        
        
    def save_data(self):
        '''
        save csv file
        '''
        
        # select file location and name
        save_file = tk.filedialog.asksaveasfilename()
        if not(save_file.endswith(".csv")):
                save_file += ".csv"

        with open(save_file, 'w') as csvFile:
            writer = csv.writer(csvFile)
            writer.writerows(self.tracks_data)

        csvFile.close()
         
        print("csv file has been saved to ", save_file)

    def read_parameters(self):
        '''
        read parameters from the GUI
        '''
        if self.param1_diameter.get()!='':
            self.maximum_diameter=int(self.param1_diameter.get())

        if self.param2_sigma.get()!='':
            self.sigma=float(self.param2_sigma.get())

        if self.param3_threshold.get()!='':
            self.threshold=float(self.param3_threshold.get())

        if self.param4_peak.get()!='':
            self.min_peak=float(self.param4_peak.get())

        if self.param5_distance.get()!='':
            self.max_dist=float(self.param5_distance.get())

        if self.param6_framegap.get()!='':
            self.frame_gap=float(self.param6_framegap.get())
            
        if self.param6_framegap.get()!='':
            self.min_track_length=float(self.param7_framegap.get())
                    
            
        if self.res_parameter.get()!='':
            self.img_resolution=float(self.res_parameter.get())
        
        if self.frame_parameter.get()!='':
            self.frame_rate=float(self.frame_parameter.get())

    def processing(self):
        '''
        preprocessing step - connected to a button
        '''
        self.movie_processed = background_substraction(self.movie.copy())

        # show the frame in the monitor
        self.show_frame()

    def preview(self):
        '''
        show random frame with detection on the monitor - connected to a button
        '''
        
        #read parameters
        self.read_parameters()

        # select movie frame
        image = self.movie_processed[self.frame_pos,:,:]

        # invert image in case you are looking at the dark spot
        if self.spot_switch==0:
            image_for_process=skimage.util.invert(image)
        else:
            image_for_process=image

        # set detector
        detect_particle=Detectors()

        #MSSEF settings
        detect_particle.c=self.threshold  # coef for the thresholding
        detect_particle.sigma=self.sigma # max sigma for LOG
        detect_particle.expected_size=self.maximum_diameter # field for gaussian fitting

        #thresholding settings
        detect_particle.min_distance=5 # minimum distance between two max after MSSEF
        detect_particle.threshold_rel=self.min_peak # min peak value in relation to the image

        #run detector
        centers=detect_particle.detect(image_for_process)

        #plot the result
        self.show_frame(centers)


    def select_movie(self):
        '''
        select movie for processing - connected to a button
        '''

        filename = tk.filedialog.askopenfilename()
        root.update()
        self.movie_file=filename

        # read the file
        self.movie=skimage.io.imread(self.movie_file)
        self.movie_length=self.movie.shape[0]
        try:
            self.lbl1.destroy()
        except:
            pass
        self.lbl1 = tk.Label(master=root, text="movie file: "+self.movie_file.split("/")[-1], bg='white')
        self.lbl1.grid(row=1, column=1, columnspan=3, pady=5, padx=5)

        
        # copy proccessed
        self.movie_processed=self.movie.copy()
        
        # show first frame in the monitor
        self.show_frame()
        
        
        def show_values(v):
            self.frame_pos=int(v)
            self.show_frame() 
          
        self.scale_movie = tk.Scale(root, from_=0, to=self.movie_processed.shape[0]-1, tickinterval=int(self.movie_processed.shape[0]/5), length=self.scale_length, width=10, orient="horizontal", command=show_values)
        self.scale_movie.set(self.frame_pos)        
        self.scale_movie.grid(row=21, column=1, columnspan=3, pady=5, padx=5) 

    def tracking(self):
        '''
        detection and linking from the selected movie - connected to a button
        '''
        print("tracker is running ...")
        
        #read parameters
        self.read_parameters()
        

        def track_to_frame(data):

            # change data arrangment from tracks to frames

            track_data_framed={}
            track_data_framed.update({'frames':[]})

            for n_frame in range(0, self.movie_processed.shape[0]):


                frame_dict={}
                frame_dict.update({'frame': n_frame})
                frame_dict.update({'tracks': []})

                #rearrange the data
                for track in data:
                    p=data.get(track)
                    if n_frame in p['frames']: # if the frame is in the track
                        frame_index=p['frames'].index(n_frame) # find position in the track

                        new_trace=p['trace'][0:frame_index+1] # copy all the traces before the frame
                        frame_dict['tracks'].append({'trackID': p['trackID'], 'trace': new_trace}) # add to the list


                track_data_framed['frames'].append(frame_dict) # add the dictionary
                
            return track_data_framed


        def save_movie(tracks, save_file):
            '''
            save movie with tracks
            '''

            track_data_framed=track_to_frame(tracks)

            final_img_set = np.zeros((self.movie_processed.shape[0], self.movie_processed.shape[1], self.movie_processed.shape[2], 3))

            for frameN in range(0, self.movie_processed.shape[0]):

                plot_info=track_data_framed['frames'][frameN]['tracks']
                frame_img=self.movie_processed[frameN,:,:]
                # Make a colour image frame
                orig_frame = np.zeros((self.movie_processed.shape[1], self.movie_processed.shape[2], 3))

                orig_frame [:,:,0] = frame_img/np.max(frame_img)*256
                orig_frame [:,:,1] = frame_img/np.max(frame_img)*256
                orig_frame [:,:,2] = frame_img/np.max(frame_img)*256

                for p in plot_info:
                    trace=p['trace']
                    trackID=p['trackID']

                    clr = trackID % len(self.color_list)
                    if (len(trace) > 1):
                        for j in range(len(trace)-1):
                            # Draw trace line
                            point1=trace[j]
                            point2=trace[j+1]
                            x1 = int(point1[1])
                            y1 = int(point1[0])
                            x2 = int(point2[1])
                            y2 = int(point2[0])
                            cv2.line(orig_frame, (int(x1), int(y1)), (int(x2), int(y2)),
                                     self.color_list[clr], 2)

                # Display the resulting tracking frame
                cv2.imshow('Tracking', orig_frame)

                ################### to save #################
                final_img_set[frameN,:,:,:]=orig_frame


                    # save results

            final_img_set=final_img_set/np.max(final_img_set)*255
            final_img_set=final_img_set.astype('uint8')
            # skimage.io.imsave(save_file, final_img_set)
            if not(save_file.endswith(".tif") or save_file.endswith(".tiff")):
                save_file += ".tif"
            imageio.volwrite(save_file, final_img_set)
            cv2.destroyAllWindows()



        # detector settings
        detect_particle=Detectors()
        
        #MSSEF settings
        detect_particle.c=self.threshold # coef for thresholding
        detect_particle.sigma=self.sigma # max sigma for LOG

        #other settings
        detect_particle.min_distance=5 # minimum distance between two max after MSSEF
        detect_particle.threshold_rel=self.min_peak # min peak value in relation to the image
        detect_particle.expected_size=self.maximum_diameter # expected particle size


        tracker = Tracker(self.max_dist, self.frame_gap,  0)


        # frame to frame detection and linking loop
        for frameN in range(0, self.movie_processed.shape[0]):
            print('frame', frameN)
            
            #detection
            frame_img=self.movie_processed[frameN,:,:]
            if self.spot_switch==0:
                frame_img=skimage.util.invert(frame_img)

            centers=detect_particle.detect(frame_img)

            #tracking
            tracker.update(centers,  frameN)

        #add remaining tracks
        for trackN in range(0, len(tracker.tracks)):
            tracker.completeTracks.append(tracker.tracks[trackN])


        # rearrange the data for saving
         
        self.tracks_data=[]
        self.tracks_data.append(['X', 'Y', 'TrackID',
                                 't'])

        data_tracks={}
        trackID=0
        for trackN in range(0, len(tracker.completeTracks)):
            #save trajectories 
            
            #if track is long enough:
            if len(tracker.completeTracks[trackN].trace)>=self.min_track_length:
                trackID+=1
              
                # check the track for missing detections
                frames=tracker.completeTracks[trackN].trace_frame
                trace=tracker.completeTracks[trackN].trace
                pos=0
                new_frames=[]
                new_trace=[]
                for frame_pos in range(frames[0], frames[-1]+1):
                    frame=frames[pos]
                    
                    if frame_pos==frame:
                        new_frames.append(frame_pos)
                        new_trace.append(trace[pos])
                        pos=pos+1
                        
                    else:
                        new_frames.append(frame_pos)
                        frame_img=self.movie_processed[frame_pos,:,:]
                        
                        # find  particle location
                        
                        point=trace[pos] # previous frame
                        
                        # define ROI 
                        data=np.zeros((detect_particle.expected_size,detect_particle.expected_size))
            
                        #start point
                        start_x=int(point[0]-detect_particle.expected_size/2)
                        start_y=int(point[1]-detect_particle.expected_size/2)
                        
                        #end point
                        end_x=int(point[0]+detect_particle.expected_size/2)
                        end_y=int(point[1]+detect_particle.expected_size/2)
                        
                        x_0=0
                        x_1=detect_particle.expected_size
                        y_0=0
                        y_1=detect_particle.expected_size
                        
                        # define ROI coordinates
                        
                        if start_x<0:
                            start_x=0
                            x_0=int(detect_particle.expected_size/2-point[0])
                            
                        if start_y<0:
                            start_y=0
                            y_0=int(detect_particle.expected_size/2-point[1]) 
                            
                        if end_x>frame_img.shape[0]:
                            end_x=frame_img.shape[0]
                            x_1=int(frame_img.shape[0]-point[0]+detect_particle.expected_size/2)
            
                        if end_y>frame_img.shape[1]:
                            end_y=frame_img.shape[1]
                            y_1=int(frame_img.shape[1]-point[1]+detect_particle.expected_size/2)
                        
                        
                        data[x_0:x_1,y_0:y_1]=frame_img[start_x:end_x, start_y:end_y]
                        
                        # subpixel localisatopm
                        x,y=detect_particle.radialsym_centre(data)
                        
                        # check that the centre is inside of the spot            
                        if y<detect_particle.expected_size and x<detect_particle.expected_size and y>=0 and x>=0:               
                            new_trace.append([x+int(point[0]-detect_particle.expected_size/2),y+int(point[1]-detect_particle.expected_size/2)])
   
                        else: # if not use the previous point
                            new_trace.append(trace[pos])                
                
                
                for pos in range(0, len(new_trace)):
                    point=new_trace[pos]
                    frame=new_frames[pos]
                    self.tracks_data.append([ point[1]*self.img_resolution, point[0]*self.img_resolution, trackID, frame*self.frame_rate])

                #save for plotting tracks
                data_tracks.update({tracker.completeTracks[trackN].track_id:{
                        'trackID':trackID,
                        'trace': new_trace,
                        'frames': new_frames,
                        'skipped_frames': 0
                        }})

        save_file = tk.filedialog.asksaveasfilename(title="save the movie with trajectories")
        if save_file:
            save_movie(data_tracks, save_file)


    def close_app(self):
        '''
        quit all the proccesses while closing the GUI
        '''
        self.quit()


def main():
    global root
    root = tk.Tk()
    MainVisual(root)
    root.mainloop()

if __name__ == "__main__":
    main()



import sys

import pandas as pd

from PyQt5 import QtCore, uic, QtWidgets
from PyQt5.QtWidgets import QMainWindow, QApplication, QDialog, QFileDialog, QMessageBox

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

from trait2d.analysis import Track, ModelDB
from trait2d.analysis.models import ModelBrownian, ModelConfined, ModelHop
from trait2d.exceptions import *

import trait2d_analysis_gui.tab.msd
import trait2d_analysis_gui.tab.adc

import os

class MainWindow(QMainWindow):

    sigTrackLoaded = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        super(MainWindow, self).__init__(*args, **kwargs)

        # Load .ui file
        dirname = os.path.dirname(__file__)
        uic.loadUi(os.path.join(dirname, 'gui_analysis.ui'), self)

        # Add widgets to tabs
        wid_msd = trait2d_analysis_gui.tab.msd.widgetMSD(self)
        wid_adc = trait2d_analysis_gui.tab.adc.widgetADC(self)

        self.layoutMSD.addWidget(wid_msd, 0, 0)
        self.layoutADC.addWidget(wid_adc, 0, 0)

        self.statusBar().setSizeGripEnabled(False)

        # Initialize with no track
        self.track = None

        self.pushButtonLoadTrack.clicked.connect(self.load_track_dialog)

        ModelDB().add_model(ModelBrownian)
        ModelDB().add_model(ModelConfined)
        ModelDB().add_model(ModelHop)

    def show_trackid_dialog(self):
        Dialog = QDialog()
        dirname = os.path.dirname(__file__)
        uic.loadUi(os.path.join(dirname, "dialog_trackid.ui"), Dialog)
        Dialog.show()
        resp = Dialog.exec_()

        if resp == QDialog.Accepted:
            return Dialog.lineEditId.text()

    def show_import_dialog(self, column_names):
        Dialog = QDialog()
        dirname = os.path.dirname(__file__)
        uic.loadUi(os.path.join(dirname, "dialog_import.ui"), Dialog)
        Dialog.show()

        Dialog.comboBoxColId.addItem('')
        for c in column_names:
            Dialog.comboBoxColX.addItem(c)
            Dialog.comboBoxColY.addItem(c)
            Dialog.comboBoxColT.addItem(c)
            Dialog.comboBoxColId.addItem(c)

        Dialog.comboBoxColX.setCurrentText('')
        Dialog.comboBoxColY.setCurrentText('')
        Dialog.comboBoxColT.setCurrentText('')
        Dialog.comboBoxColId.setCurrentText('')

        if 'x' in column_names:
            Dialog.comboBoxColX.setCurrentText('x')
        if 'y' in column_names:
            Dialog.comboBoxColY.setCurrentText('y')
        if 't' in column_names:
            Dialog.comboBoxColT.setCurrentText('t')
        if 'id' in column_names:
            Dialog.comboBoxColId.setCurrentText('id')

        resp = Dialog.exec()

        if resp == QDialog.Accepted:
            return  Dialog.comboBoxLength.currentText(), \
                    Dialog.comboBoxTime.currentText(), \
                   [Dialog.comboBoxColX.currentText(),
                    Dialog.comboBoxColY.currentText(),
                    Dialog.comboBoxColT.currentText(),
                    Dialog.comboBoxColId.currentText()]

        return None, None, None

    def load_track(self, filename : str, id=None, unit_length = None, unit_time = None, col_names = None):
        if unit_length is None or unit_time is None or col_names is None:
            unit_length, unit_time, col_names = self.show_import_dialog(pd.read_csv(filename).columns)
            if unit_length is None or unit_time is None or col_names is None:
                return
        try:
            self.track = Track.from_file(filename, id=id, unit_length=unit_length, unit_time=unit_time, col_name_x = col_names[0], col_name_y = col_names[1], col_name_t = col_names[2], col_name_id = col_names[3]).normalized(normalize_t = True, normalize_xy = False)
        except LoadTrackMissingIdError:
            id = self.show_trackid_dialog()
            if id is None:
                return
            self.load_track(filename=filename, id=id, unit_length=unit_length, unit_time=unit_time, col_names=col_names)
        except LoadTrackIdNotFoundError:
            mb = QMessageBox()
            mb.setText("There is no track with that ID!")
            mb.setWindowTitle("Error")
            mb.setIcon(QMessageBox.Critical)
            mb.exec()
        except KeyError as e:
            mb = QMessageBox()
            mb.setText("Key not found: {}!".format(str(e)))
            mb.setWindowTitle("Error")
            mb.setIcon(QMessageBox.Critical)
            mb.exec()
        if self.track==None:
            return

        self.sigTrackLoaded.emit()

        self.statusbar.showMessage(
            "Loaded track of length {}".format(self.track.get_size()))      

    def load_track_dialog(self):
        filename, _ = QFileDialog.getOpenFileName(
            self, "Load track file", "", "Track files (*.csv)")
        if filename == '':
            return
        self.load_track(filename, id=None)


def main():
    app = QApplication(sys.argv)

    main = MainWindow()
    main.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



from trait2d.analysis import ModelDB

import numpy as np

def delete_adc_analysis_results(self):
    """ Delete the ADC analysis results."""
    self._adc_analysis_results = None

def get_adc_analysis_results(self):
    """Returns the ADC analysis results."""
    return self._adc_analysis_results

def adc_analysis(self, R: float = 1/6, fraction_fit_points: float=0.25, fit_max_time: float = None, maxfev = 1000, enable_log_sampling = False, log_sampling_dist = 0.2, weighting = 'error'):
    """Revised analysis using the apparent diffusion coefficient

    Parameters
    ----------
    R: float
        Point scanning across the field of view.
    fraction_fit_points: float
        Fraction of points to use for fitting. Defaults to 25 %.
    fit_max_time: float
        Maximum time in fit range. Will override fraction_fit_points.
    maxfev: int
        Maximum function evaluations by scipy.optimize.curve_fit. The fit will fail if this number is exceeded.
    enable_log_sampling: bool
        Only sample logarithmically spaced time points for analysis.
    log_sampling_dist: float
        Exponent of logarithmic sampling (base 10).
    weighting: str
        Weighting of the datapoints used in the fit residual calculation. Can be `error` (weight by inverse standard
        deviation), `inverse_variance` (weight by inverse variance), `variance` (weight by variance)
        or `disabled` (no weighting). Default is `error`.

    Returns
    -------
    adc_analysis_results: dict
        Dictionary containing all analysis results.
        Can also be retreived using `Track.get_adc_analysis_results()`.
    """
    # Calculate MSD if this has not been done yet.
    if self._msd is None:
        self.calculate_msd()

    dt = self._t[1] - self._t[0]

    N = self._msd.size

    # Time coordinates
    # This is the time array, as the fits will be MSD vs T
    T = np.linspace(dt, dt*N, N, endpoint=True)

    # Compute  the time-dependent apparent diffusion coefficient.
    Dapp = self._msd / (4 * T * (1 - 2*R*dt / T))
    Dapp_err = self._msd_error / (4 * T * (1 - 2*R*dt / T))

    model, fit_indices, fit_results = self._categorize(np.array(Dapp), np.arange(
        1, N+1), Dapp_err = Dapp_err, R=R, fraction_fit_points=fraction_fit_points, fit_max_time=fit_max_time, maxfev=maxfev, enable_log_sampling=enable_log_sampling, log_sampling_dist=log_sampling_dist, weighting = weighting)

    self._adc_analysis_results = {}
    self._adc_analysis_results["Dapp"] = np.array(Dapp)
    self._adc_analysis_results["Dapp_err"] = np.array(Dapp_err)
    self._adc_analysis_results["fit_indices"] = fit_indices
    self._adc_analysis_results["fit_results"] = fit_results
    self._adc_analysis_results["best_model"] = model

    return self._adc_analysis_results

    
def plot_adc_analysis_results(self):
    """Plot the ADC analysis results.

    Raises
    ------
    ValueError
        Track has not been analyzed using ADC analysis yet.
    """
    import matplotlib.pyplot as plt
    if self.get_adc_analysis_results() is None:
        raise ValueError(
            "Track has not been analyzed using adc_analysis yet!")

    dt = self._t[1] - self._t[0]
    N = self._t.size
    T = np.linspace(1, N, N-3,  endpoint=True) * dt

    fit_results = self.get_adc_analysis_results()["fit_results"]

    Dapp = self.get_adc_analysis_results()["Dapp"]
    Dapp_err = self.get_adc_analysis_results()["Dapp_err"]
    idxs = self.get_adc_analysis_results()["fit_indices"]

    n_points = idxs[-1]

    plt.figure(figsize=(8, 4))
    plt.grid(linestyle='dashed', color='grey')
    plt.semilogx(T, Dapp, label="Data", color='black')
    plt.fill_between(T, Dapp-Dapp_err, Dapp+Dapp_err, color='black', alpha=0.5)
    plt.semilogx(T[idxs], Dapp[idxs], label="Sampled Points", linestyle="", marker="|", markersize=15.0, color='red', zorder=-1)
    for model in fit_results:
        r = fit_results[model]["params"]
        rel_likelihood = fit_results[model]["rel_likelihood"]
        ks_p_value = fit_results[model]["KStestPValue"]
        m = None
        for c in ModelDB().models:
            if c.__class__.__name__ == model:
                m = c
                break
        if m is None:
            raise ValueError("Can't plot results for model {}; make sure the model is loaded in ModelDB()".format(model))
        pred = m(T, *r)
        plt.semilogx(T[0:n_points], pred[0:n_points],
                    label=f"{model}\nrel. likelihood={rel_likelihood:.2e}\nKS p-value={ks_p_value:.2e}")
    model = self.get_adc_analysis_results()["best_model"]

    plt.axvspan(T[0], T[n_points], alpha=0.25,
                color='gray', label="Fit region")
    plt.xlabel("Time in s")
    plt.ylabel("Normalized ADC")
    plt.title("Diffusion Category: {}".format(model))
    plt.xlim(T[0], T[-1])
    plt.legend(bbox_to_anchor=(1, 1), loc='upper left')
    plt.subplots_adjust(right=0.7)
    plt.show()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np


def background_substraction(movie, saturation = 0.3, nFrames = 1000, invert=True):
    """iScat Background substraction based on the  original thesis processing
    Parameters
    ----------
    movie : ndarray
        The movie to process, an ndarray of shape (time, row, col)
    saturation : float
        The pixel value saturation used for brightness adjustment (between 0 and 100)
    nFrames : int
        Number of frames to use at the beginning of the movie for flat-field computation
    invert : bool
        Bool to invert the intensity.

    Returns
    -------
    movie_p : ndarray
        Processed movie

    """
    dtype = movie.dtype
    movie_p = np.copy(movie).astype(np.float32)

    # Compute flat field from first 1000 frames (lateral shift)
    bg_median = np.median(movie_p[0:nFrames,...], axis=0)
    movie_p = movie_p / bg_median

    # Temporal average
    bg_avg = np.mean(movie_p, axis=0)

    # Background removal
    movie_p = movie_p - bg_avg

    # Normalization
    imin = movie_p.min()
    imax = movie_p.max()
    movie_p = (movie_p - imin) / (imax - imin)

    # Brightness adjustment
    isat = np.percentile(movie_p, 100 - saturation)
    movie_p = movie_p / isat
    movie_p[movie_p>1] = 1

    # Inverting the contrast
    if invert:
        movie_p = 1 - movie_p

    # Convert back to the original pixel format
    movie_p = (np.iinfo(dtype).max * movie_p).astype(dtype)

    return movie_p



#!/usr/bin/env python3
# coding: utf-8
import numpy as np
import warnings
from scipy import spatial
from skimage import draw
import matplotlib.pyplot as plt
import tqdm
import json
import csv
import pprint
from pathlib import Path
import pickle
from skimage import util as sk_util
from scipy.signal import fftconvolve
import imageio as io
from pathlib import Path

DEBUG = True


# Abstract diffusion object
class Diffusion(object):
    """Abstract diffusion simulator

    The methods to modify after inherition are: self.run, and te self.params_list, and assign the input variables
    """
    def __init__(self, Tmax=1.0, dt=1e-3, L=1, dL=1e-3, d=1e-2, seed: int=None, quantize=True):

        ## Initializations
        self.params_list = ["Tmax", "dt", "L", "dL", "d", "seed", "quantize"]
        self.Tmax = Tmax
        self.dt = dt
        self.L = L
        self.dL = dL
        self.d = d
        self.seed = seed
        self.quantize = quantize

        # Set random seed
        np.random.seed(self.seed)

    def run(self):
        # Initialize
        self.trajectory = dict()
        self.trajectory["x"] = []
        self.trajectory["y"] = []
        self.trajectory["t"] = []
        self.trajectory["id"] = []

    def display_trajectory(self, time_resolution=None, limit_fov=False, alpha=0.8, title="Diffusion"):
        """Display the simulated trajectory.

        Parameters
        ----------

        param time_resolution:
            [s]
        """
        assert hasattr(self, "trajectory"), "You must first run the simulator"
        if time_resolution is not None:
            time_interval = int(np.ceil(time_resolution / self.dt))
        else:
            time_interval = 1
        x = np.array(self.trajectory["x"][0::time_interval])
        y = np.array(self.trajectory["y"][0::time_interval])
        plt.plot(x, y, alpha=alpha)

        if limit_fov:
            plt.xlim((0, self.parameters["L"]))
            plt.ylim((0, self.parameters["L"]))

        plt.title(title)
        #plt.show()

    def _gather_parameters(self):
        self.parameters = dict()
        for this_param in self.params_list:
            self.parameters[this_param] = getattr(self, this_param)

    def _set_parameters(self, parameters):
        # Parameters must be a dict
        for this_param in self.params_list:
            if this_param in parameters:
                setattr(self, this_param, parameters[this_param])

        # Set the random seed if given
        if "seed" in parameters:
            np.random.seed(parameters["seed"])

    def print_parameters(self):
        """Print the simulation parameters, accessible from self._gather_parameters()
        """
        self._gather_parameters()
        pprint.pprint(self.parameters)

    def save_trajectory(self, filename, format=None):
        """Save the simulated trajectory as either a json, cvs or pcl file with fiels t, x, and y
        """
        supported_formats = ["json", "csv", "pcl"]
        if format is None:
            format = Path(filename).suffix.replace(".","")
        assert hasattr(self, "trajectory"), "You must first run the simulator to obtain a trajectory"
        assert format in supported_formats, f"Supported formats are: {supported_formats}"

        if format == "json":
            with open(filename, "w") as f:
                json.dump(self.trajectory, f)
        elif format == "csv":
            with open(filename, "w", newline='') as f:
                fieldnames = ["t", "x", "y", "id"]
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                writer.writeheader()
                for i in range(len(self.trajectory["x"])):
                    this_row = {"t": self.trajectory["t"][i],
                                "x": self.trajectory["x"][i],
                                "y": self.trajectory["y"][i],
                               "id": 0}
                    writer.writerow(this_row)

        elif format == "pcl":
            with open(filename, "wb") as f:
                pickle.dump(self.trajectory, f)

    def save_parameters(self, filename, mkdir=True):
        """Save the simulation parameters, accessible through _gather_parameter
        """
        self._gather_parameters()

        # Create the parent directory
        if mkdir:
            Path(filename).parent().mkdir(parents=True, exist_ok=True)

        # Save the parameters as a JSON file
        with open(filename, "w") as f:
            json.dumps(f, self.parameters)

    def load_parameters(self, filename):
        """Load the simulation parameters
        """
        # Load the parameters from a JSON file
        with open(filename, 'r') as f:
            parameters = json.load(f)

        # Set the loaded parameters
        self._set_parameters(parameters)

# Brownian diffusion simulation
class BrownianDiffusion(Diffusion):
    """BrownianDiffusion Initialization

    Parameters
    ----------
    Tmax: float
        Maximum simulation time (s)
    dt: float
        Simulation time resolution (s)
    L: float
        Simulation domain size (m)
    dL: float
        Simulation spatial resolution (m)
    d: float
        Diffusion coefficient (m^2/s)
    seed: int
        Seed to initialize the random generator (for reproducibility)
    quantize: bool
        Quantize the position to the simulation spatial resolution grid.
    """
    def __init__(self, **kwargs):
        super(BrownianDiffusion, self).__init__(**kwargs)

    def run(self):
        """Run a random walk simulation (Brownian diffusion)
        """
        # Starting values is in the center.
        x = self.L / 2.0
        y = self.L / 2.0
        diffuse = True
        iteration = 0

        x_list = [x]
        y_list = [y]
        t_list = [0]

        pbar = tqdm.tqdm(desc="Brownian Diffusion Simulation", total=int(self.Tmax / self.dt))
        while iteration * self.dt < self.Tmax and diffuse:
            # Update position and compartment
            x0 = x + np.sqrt(self.d * self.dt) * np.random.randn()
            y0 = y + np.sqrt(self.d * self.dt) * np.random.randn()

            # Outside of area, stop it here
            if not(0 <= x0 <= self.L) or not(0 <= y0 <= self.L):
                diffuse = False
                continue

            # Update the position and compartment
            x = x0
            y = y0
            x_list.append(x)
            y_list.append(y)
            t_list.append(t_list[-1]+self.dt)
            iteration += 1
            pbar.update()
        pbar.close()
        print(f"Free diffusion simulation was completed in {iteration} iterations.")

        # Make sure the simulated track is expressed as dL spatial resolution
        if self.quantize:
            x_list = (np.round(np.array(x_list) / self.dL) * self.dL).tolist()
            y_list = (np.round(np.array(y_list) / self.dL) * self.dL).tolist()

        self.trajectory = {"x": x_list, "y": y_list, "t": t_list}

# Hopping diffusion simulation
class HoppingDiffusion(Diffusion):
    """Simulates a hopping diffusion trajectory of a single molecule for a
    hopping diffusion model (i.e. free diffusion inside of compartments but
    changing from one compartment to the next only with a certain probability)

    **Syntax**:

    .. code-block:: python

        simulator = hopping_trajectories(Tmax, dt, L, Df, HL, HP, seed)
        simulator.run()

    **Authors**:

    | Jan Keller-Findeisen (jkeller1@gwdg.de)
    | Department of NanoBiophotonics
    | Max Planck Institute of Biophysical Chemistry
    | Am Fassberg 11, 3077 Goettingen, Germany

    Converted to Python by Joel Lefebvre

    Parameters
    ----------
    Tmax : float
        Maximal total length [s]
    dt : float
        Time step [s]
    L : float
        Sandbox length [m]
    dL : float
        Compartment map pixel size [m]
    Df : float
        Free diffusion coefficient [m^2/s]
    HL : float
        Average compartment diameter/length [m]
    HP : float
        Hopping probability [0-1]
    seed : float
        Random generator seed (nonnegative integer)
    quantize: bool
        Quantize the position to the simulation spatial resolution grid.
    """
    def __init__(self, Tmax=100, dt=50e-6, L=10e-6, dL=20e-9, Df=8e-13, HL=40e-9, HP=0.01, seed: int = None, quantize=True):

        # super(HoppingDiffusion, self).__init__(seed=seed)

        # Set parameters
        self.params_list = ["Tmax", "dt", "L", "dL", "Df", "HL", "HP", "seed", "quantize"]
        self.Tmax = Tmax
        self.dt = dt
        self.L = L
        self.dL = dL
        self.Df = Df
        self.HL = HL
        self.HP = HP
        self.seed = seed
        self.quantize = quantize

        # Set random seed
        np.random.seed(self.seed)

        # Mean displacement in two directions
        self.md = np.sqrt(4 * self.Df * self.dt)
        if self.md > self.HL / 2.0:
            warnings.warn("mean displacement (2D) is more than half of average compartment length, results might be compromised")
        self.d = self.md / np.sqrt(2)  # we need the mean displacement in 1D later

        # Number of compartments
        self.num = int(np.round((self.L / self.HL)**2))

    def create_hopping_map(self):
        ## Create hopping map
        centers = self.L * np.random.rand(self.num, 2)

        # Compute Voronoi diagram
        vor = spatial.Voronoi(centers)

        # Fill map with ids
        xm, ym = np.meshgrid(np.arange(0, self.L, self.dL), np.arange(0, self.L, self.dL))
        m = np.zeros(xm.shape)
        id = 0
        nc = len(vor.regions)
        for kc in range(nc):
            v = vor.regions[kc]
            v = [x for x in v if x != -1]
            if len(v) > 0:
                v.append(v[0])
                p = np.round(vor.vertices[v] / self.dL)
                id += 1
                # cv2.polylines(m, v, isClosed=True, color=id)
                rr, cc = draw.polygon(p[:,0], p[:,1], shape=xm.shape)
                m[rr,cc] = id

        # Add the hopping map to the internals
        self.hopping_map = m
        self.centers = centers
        self.voronoi_vertices = vor.vertices
        self.voronoi_regions = vor.regions

    def run(self, update_hoppingMap=False):
        ## Create hopping trajectory

        # Create or update the hopping map
        if not hasattr(self, "hopping_map") or update_hoppingMap:
            self.create_hopping_map()

        # Starting values, in the center of the map
        x = self.L / 2.0
        y = self.L / 2.0
        id = self.hopping_map[int(np.floor(x/self.dL)), int(np.floor(y/self.dL))]
        diffuse = True
        iteration = 0

        buffer = np.zeros((int(1e5), 3))
        buffer_counter = 0
        store = dict()
        store_counter = 0
        t = 0

        pbar = tqdm.tqdm(desc="Simulation", total=int(self.Tmax / self.dt))
        while iteration * self.dt < self.Tmax and diffuse: # the indefinitely long loop for each time step
            # Update position and compartment
            x0 = x + self.d * np.random.randn()
            y0 = y + self.d * np.random.randn()

            # Outside of area, stop it here
            if x0 < 0 or x0 > self.L or y0 < 0 or y0 > self.L:
                diffuse = False
                continue

            # Get the new position id
            id0 = self.hopping_map[int(np.floor(x0/self.dL)), int(np.floor(y0/self.dL))]

            # Different compartment, and we do not hop
            if id0 != id and np.random.rand() > self.HP:
                # Retry until compartment is the same again
                while id0 != id:
                    x0 = x + self.d * np.random.randn()
                    y0 = y + self.d * np.random.randn()
                    ix = int(np.floor(x0 / self.dL))
                    iy = int(np.floor(y0 / self.dL))
                    if 0 <= ix < self.hopping_map.shape[0] and 0 <= iy < self.hopping_map.shape[1]:
                        id0 = self.hopping_map[ix,iy]

            # Update the position and compartment
            x = x0
            y = y0
            id = id0
            iteration += 1

            # Store position and compartment
            buffer[buffer_counter,:] = [x, y, id]
            buffer_counter += 1

            if buffer_counter >= buffer.shape[0]:
                store[store_counter] = buffer
                store_counter += 1
                buffer = np.zeros((int(1e5), 3))
                buffer_counter = 0

            t += self.dt
            pbar.update()
            if t > self.Tmax:
                diffuse = False
        pbar.close()
        print(f"Hopping diffusion simulation was completed in {iteration} iterations.")

        # Trim buffer
        buffer = buffer[0:buffer_counter,:]
        store[store_counter] = buffer

        # Gather the simulation results
        x_list = []
        y_list = []
        id_list = []
        for key in store.keys():
            x_list.extend(list(store[key][:, 0]))
            y_list.extend(list(store[key][:, 1]))
            id_list.extend(list(store[key][:, 2]))

        # Quantize the position
        if self.quantize:
            x_list = (np.round(np.array(x_list) / self.dL) * self.dL).tolist()
            y_list = (np.round(np.array(y_list) / self.dL) * self.dL).tolist()

        t_list = list(np.linspace(0, len(x_list) * self.dt, len(x_list)))
        self.trajectory = dict()
        self.trajectory["x"] = x_list
        self.trajectory["y"] = y_list
        self.trajectory["t"] = t_list
        self.trajectory["id"] = id_list

    def display_hopping_map(self):
        assert hasattr(self, "hopping_map"), "No hopping map was set or created."
        plt.imshow(self.hopping_map)
        plt.title("Hopping Map")

# iScat Acquisition simulation
class iscat_movie(object):
    # Notes
    # TODO: Add PSF & Object shape inputs (instead of only psf)
    # TODO: Add z-phase jitter for the PSF instead of using a fixed plane
    # TODO: Load a simulation parameters file instead of passing everything in the command line
    # TODO: Use input size as alternative
    # TODO: link tqdm with logging
    # TODO: Create a python wrapper for the ImageJ plugin 'DeconvolutionLab2' to generate PSF in the script?
    # TODO: Background noise with different statistics (similar to transcient particles)
    def __init__(self, tracks=None, resolution=1.0, dt=1, snr=25, background=0.3, noise_gaussian=0.15, noise_poisson=True, ratio="square"):
        # Prepare the simulator
        self.resolution = resolution
        self.snr = snr  # Signal to noise ratio
        self.background = background  # Background intensity
        self.noise_gaussian = noise_gaussian  # Gaussian noise variance
        self.noise_poisson = noise_poisson  # Poisson noise variance
        self.dt = dt  # Temporal resolution
        self.ratio = ratio

        if isinstance(tracks, dict):
            self.tracks = tracks
        elif isinstance(tracks, str) or isinstance(tracks, Path):
            self.load_tracks(tracks)

    def initialize(self):
        assert hasattr(self, 'tracks'), "You must load a tracks file or set a tracks dict first"
        self.n_spots = len(self.tracks["x"])

        # Get the number of frames
        self.tmin = 0
        self.tmax = np.max(self.tracks["t"])
        self.n_frames = int((self.tmax - self.tmin)/self.dt) + 1


        # Get the movie shape
        self.xmin = np.min(self.tracks["x"])
        self.ymin = np.min(self.tracks["y"])
        self.xmax = np.max(self.tracks["x"])
        self.ymax = np.max(self.tracks["y"])
        if self.ratio == "square":
            self.xmin = min(self.xmin, self.ymin)
            self.ymin = min(self.xmin, self.ymin)
            self.xmax = max(self.xmax, self.ymax)
            self.ymax = max(self.xmax, self.ymax)
        else:
            print(f"Unknown ratio: {self.ratio}")

        # Initialize the simulation grid
        x = np.linspace(self.xmin, self.xmax, int((self.xmax - self.xmin) / self.resolution))
        y = np.linspace(self.ymin, self.ymax, int((self.ymax - self.ymin) / self.resolution))
        self.nx = len(x) + 1
        self.ny = len(y) + 1
        
        print(" - - - - - - - - ")
        print(self.tmax)
        print(len(x))
        print("self.xmin", self.xmin)
        print("self.ymin", self.ymin)
        print("self.xmax", self.xmax)

        print(f"Movie shape will be: ({self.nx}, {self.ny}) with ({self.n_frames}) frames")

    def run(self):
        self.initialize()

        # Create the movie array
        print("Creating an empty movie")
                

        
        movie = np.ones((self.n_frames, self.nx, self.ny)) * self.background

        # Add Gaussian noise to the background
        if self.noise_gaussian is not None:
            print("Adding gaussian noise to the background")
            movie = sk_util.random_noise(movie, mode="gaussian", var=self.noise_gaussian)

        # Populate the tracks
        for this_spot in tqdm.tqdm(range(self.n_spots), "Adding tracks"):
            mx = int(np.round((self.tracks['x'][this_spot] - self.xmin) / self.resolution))
            my = int(np.round((self.tracks['y'][this_spot] - self.ymin) / self.resolution))
            mt = int(self.tracks["t"][this_spot] / self.dt)
            if isinstance(mx, list):
                for x, y, t in zip(mx, my, mt):
                    if (0 <= mx < self.nx) and (0 <= my < self.ny):
                        movie[t, x, y] = movie[t, x, y] + (1 + self.snr) * self.background
            else:
                if (0 <= mx < self.nx) and (0 <= my < self.ny):
                    movie[mt, mx, my] = movie[mt, mx, my] + (1 + self.snr) * self.background

        # Convolve by PSF if provided
        if hasattr(self, "psf_2d"):
            px, py = self.psf_2d.shape
            movie = np.pad(movie, ((0, 0), (px // 2, px // 2), (py // 2, py // 2)), mode="reflect")

            # Apply convolution
            for i in tqdm.tqdm(range(self.n_frames), desc="Convolving with PSF"):
                movie[i, ...] = fftconvolve(movie[i, ...], self.psf_2d, mode='same')

            # Unpad
            movie = movie[:, px // 2:px // 2 + self.nx, py // 2:py // 2 + self.ny]

        # Add Poisson noise
        if self.noise_poisson:
            print("Adding Poisson noise")
            movie = sk_util.random_noise(movie, mode="poisson", clip=False)
            movie[movie < 0] = 0

        self.movie = movie

    def save(self, filename):
        assert hasattr(self, "movie"), "You must first run the simulation"
        io.volwrite(filename, self.movie.astype(np.float32))


    def load_tracks(self, filename, field_x="x", field_y="y", field_t="t", field_id="id", file_format=None): # TODO: Load other tracks format
        """Load the tracks from a csv file.
        
        Parameters
        ----------
        filename : str
            Path to a csv filename
        field_x : str
            Column name in the CSV corresponding to the tracks X positions.
        field_y : str
            Column name in the CSV corresponding to the tracks Y positions.
        field_t : str
            Column name in the CSV corresponding to the tracks time.
        field_id : str
            Column name in the CSV corresponding to the tracks ID.
        file_format : str
            Specify the file format (available are cvs, json, pcl). If none is given, it will be inferred from the filename
        """
        tracks = {"x": [], "y": [], "t": [], "id": []}
        if Path(filename).suffix == ".csv" or file_format == "csv":
            # Load the csv file
            with open(filename, "r") as csvfile:
                #  Detect the csv format
                dialect = csv.Sniffer().sniff(csvfile.read())

                #  Create a reader
                csvfile.seek(0)
                reader = csv.reader(csvfile, dialect)

                for i, row in enumerate(reader):
                    if i == 0:
                        column_names = row
                    else:
                        tracks["x"].append(float(row[column_names.index(field_x)]))
                        tracks["y"].append(float(row[column_names.index(field_y)]))
                        tracks["t"].append(float(row[column_names.index(field_t)]))
                        tracks["id"].append(int(row[column_names.index(field_id)]))
        elif Path(filename).suffix == ".json" or file_format == "json":
            with open(filename, "r") as f:
                content = json.load(f)
            tracks["x"] = content[field_x]
            tracks["y"] = content[field_y]
            tracks["t"] = content[field_t]
            tracks["id"] = content[field_id]

        elif Path(filename).suffix == ".pcl" or file_format == "pcl":
            with open(filename, "rb") as f:
                content = pickle.load(f)
            tracks["x"] = content[field_x]
            tracks["y"] = content[field_y]
            tracks["t"] = content[field_t]
            tracks["id"] = content[field_id]

        self.tracks = tracks

    def load_psf(self, filename):
        psf = io.volread(filename).squeeze()
        psf_2d = psf[int(psf.shape[0] / 2), ...].squeeze()
        psf_2d = psf_2d / psf_2d.sum()
        self.psf_2d = psf_2d



from trait2d.analysis import Track, ModelDB
from trait2d.analysis.models import ModelBrownian, ModelConfined, ModelHop
from trait2d.exceptions import *

import numpy as np

from PyQt5 import uic
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMessageBox, QWidget, QApplication

from trait2d_analysis_gui.plot import ModelFitWidget
from trait2d_analysis_gui.render_math import MathTextLabel

import os

class widgetADC(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self)
        
        dirname = os.path.dirname(__file__)
        uic.loadUi(os.path.join(dirname, 'adc.ui'), self)

        self.parent = parent

        self.pushButtonAnalyze.clicked.connect(self.analyze)
        self.checkBoxLogPlot.stateChanged.connect(self.plot.set_log)
        self.pushButtonClipboard.clicked.connect(self.results_to_clipboard)
        self.pushButtonSetRange.clicked.connect(self.set_range_from_spinbox)

        self.parent.sigTrackLoaded.connect(self.reset)
        self.plot.sigFitRangeChanged.connect(self.on_range_changed)
        self.plot.setLabel('left', "Dapp")
        self.plot.setLabel('bottom', "T", units="s")

        self.legend = self.plot.addLegend()
        self.plot_dapp = self.plot.plot([], name='Dapp')
        self.plot_brownian = self.plot.plot([], pen=(1, 3), name = 'Brownian')
        self.plot_confined = self.plot.plot([], pen=(2, 3), name = 'Confined')
        self.plot_hopping  = self.plot.plot([], pen=(3, 3), name = 'Hopping')

    def results_to_clipboard(self):
        if self.parent.track is None or self.parent.track.get_adc_analysis_results() is None:
            self.parent.statusbar.showMessage("Analyze a track first!")
            mb = QMessageBox()
            mb.setText("Analyze a track first!")
            mb.setWindowTitle("Error")
            mb.setIcon(QMessageBox.Warning)
            mb.exec()
            return

        QApplication.clipboard().setText(str(self.parent.track.get_adc_analysis_results()["fit_results"]))

    def reset(self):
        self.lineEditParam1_1.setText("")
        self.lineEditParam2_1.setText("")
        self.lineEditRelLikelihood_1.setText("")
        self.lineEditBIC_1.setText("")

        self.lineEditParam1_2.setText("")
        self.lineEditParam2_2.setText("")
        self.lineEditParam3_2.setText("")
        self.lineEditRelLikelihood_2.setText("")
        self.lineEditBIC_2.setText("")

        self.lineEditParam1_3.setText("")
        self.lineEditParam2_3.setText("")
        self.lineEditParam3_3.setText("")
        self.lineEditParam4_3.setText("")
        self.lineEditRelLikelihood_3.setText("")
        self.lineEditBIC_3.setText("")

        self.plot.reset()

    def analyze(self):
        if self.parent.track == None:
            self.parent.statusbar.showMessage("Load a track first!")
            mb = QMessageBox()
            mb.setText("Load a track first!")
            mb.setWindowTitle("Error")
            mb.setIcon(QMessageBox.Warning)
            mb.exec()
            return

        T = self.parent.track.get_t()[0:-3]
        fit_max_time = self.plot.get_range()
        if fit_max_time <= 0.0:
            fit_max_time = None

        maxfev = int(self.lineEditMaxIt.text())

        model_hop = ModelDB().get_model(ModelHop)
        model_brownian = ModelDB().get_model(ModelBrownian)
        model_confined = ModelDB().get_model(ModelConfined)

        if self.checkBoxUseInitial.checkState():
            if (self.lineEditParam1_1.text() != ""):
                model_brownian.initial[0] = float(self.lineEditParam1_1.text())
            if (self.lineEditParam2_1.text() != ""):
                model_brownian.initial[1] = float(self.lineEditParam2_1.text())

            if (self.lineEditParam1_2.text() != ""):
               model_confined.initial[0] = float(self.lineEditParam1_2.text())
            if (self.lineEditParam2_2.text() != ""):
                model_confined.initial[1] = float(self.lineEditParam2_2.text())
            if (self.lineEditParam3_2.text() != ""):
                model_confined.initial[2] = float(self.lineEditParam3_2.text())

            if (self.lineEditParam1_3.text() != ""):
                model_hop.initial[0] = float(self.lineEditParam1_3.text())
            if (self.lineEditParam2_3.text() != ""):
                model_hop.initial[1] = float(self.lineEditParam2_3.text())
            if (self.lineEditParam3_3.text() != ""):
                model_hop.initial[2] = float(self.lineEditParam3_3.text())
            if (self.lineEditParam4_3.text() != ""):
                model_hop.initial[3] = float(self.lineEditParam4_3.text())

        R = float(self.doubleSpinBoxInputParam1.value())

        try:
            results = self.parent.track.adc_analysis(R=R, fit_max_time=fit_max_time, maxfev=maxfev)
        except RuntimeError:
            mb = QMessageBox()
            mb.setText("A model fit failed! Try raising the maximum iterations or different initial values.")
            mb.setWindowTitle("Fit Error")
            mb.setIcon(QMessageBox.Warning)
            mb.exec()
            return

        fit_results = results["fit_results"]

        # Show results for brownian model in GUI
        self.lineEditParam1_1.setText("{:5e}".format(fit_results["ModelBrownian"]["params"][0]))
        self.lineEditParam1Error_1.setText("{:5e}".format(fit_results["ModelBrownian"]["errors"][0]))
        self.lineEditParam2_1.setText("{:5e}".format(fit_results["ModelBrownian"]["params"][1]))
        self.lineEditParam2Error_1.setText("{:5e}".format(fit_results["ModelBrownian"]["errors"][1]))
        self.lineEditRelLikelihood_1.setText("{:5f}".format(fit_results["ModelBrownian"]["rel_likelihood"]))
        self.lineEditBIC_1.setText("{:5f}".format(fit_results["ModelBrownian"]["bic"]))

        # Show results for confined model in GUI
        self.lineEditParam1_2.setText("{:5e}".format(fit_results["ModelConfined"]["params"][0]))
        self.lineEditParam1Error_2.setText("{:5e}".format(fit_results["ModelConfined"]["errors"][0]))
        self.lineEditParam2_2.setText("{:5e}".format(fit_results["ModelConfined"]["params"][1]))
        self.lineEditParam2Error_2.setText("{:5e}".format(fit_results["ModelConfined"]["errors"][1]))
        self.lineEditParam3_2.setText("{:5e}".format(fit_results["ModelConfined"]["params"][2]))
        self.lineEditParam3Error_2.setText("{:5e}".format(fit_results["ModelConfined"]["errors"][2]))
        self.lineEditRelLikelihood_2.setText("{:5f}".format(fit_results["ModelConfined"]["rel_likelihood"]))
        self.lineEditBIC_2.setText("{:5f}".format(fit_results["ModelConfined"]["bic"]))

        # Show results for hopping in GUI
        self.lineEditParam1_3.setText("{:5e}".format(fit_results["ModelHop"]["params"][0]))
        self.lineEditParam1Error_3.setText("{:5e}".format(fit_results["ModelHop"]["errors"][0]))
        self.lineEditParam2_3.setText("{:5e}".format(fit_results["ModelHop"]["params"][1]))
        self.lineEditParam2Error_3.setText("{:5e}".format(fit_results["ModelHop"]["errors"][1]))
        self.lineEditParam3_3.setText("{:5e}".format(fit_results["ModelHop"]["params"][2]))
        self.lineEditParam3Error_3.setText("{:5e}".format(fit_results["ModelHop"]["errors"][2]))
        self.lineEditParam4_3.setText("{:5e}".format(fit_results["ModelHop"]["params"][3]))
        self.lineEditParam4Error_3.setText("{:5e}".format(fit_results["ModelHop"]["errors"][3]))
        self.lineEditRelLikelihood_3.setText("{:5f}".format(fit_results["ModelHop"]["rel_likelihood"]))
        self.lineEditBIC_3.setText("{:5f}".format(fit_results["ModelHop"]["bic"]))

        dt = T[1] - T[0]

        Dapp = self.parent.track.get_adc_analysis_results()["Dapp"]
        reg1 = fit_results["ModelBrownian"]["params"]
        reg2 = fit_results["ModelConfined"]["params"]
        reg3 = fit_results["ModelHop"]["params"]
        m1 = model_brownian(T, *reg1)
        m2 = model_confined(T, *reg2)
        m3 = model_hop(T, *reg3)       

        self.plot.reset()
        self.plot.setup()

        indexes = results["fit_indices"]

        self.plot_dapp.setData(T, Dapp)
        self.plot_brownian.setData(T[indexes], m1[indexes])
        self.plot_confined.setData(T[indexes], m2[indexes])
        self.plot_hopping.setData(T[indexes], m3[indexes])

        self.plot.set_range(T[indexes[-1]])
        self.plot.autoRange()     

    def show_formula_model_1(self):
        mathText = r'$todo$'
        mb = MathTextLabel(mathText)
        mb.setFocus(True)
        mb.setWindowModality(Qt.ApplicationModal)
        mb.show()

    def show_formula_model_2(self):
        mathText = r'$todo$'
        mb = MathTextLabel(mathText)
        mb.setFocus(True)
        mb.setWindowModality(Qt.ApplicationModal)
        mb.show()

    def show_formula_model_3(self):
        mathText = r'$todo$'
        mb = MathTextLabel(mathText)
        mb.setFocus(True)
        mb.setWindowModality(Qt.ApplicationModal)
        mb.show()
    
    def on_range_changed(self):
        self.doubleSpinBoxRange.setValue(self.plot.get_range() * 1000.0)

    def set_range_from_spinbox(self):
        self.plot.set_range(self.doubleSpinBoxRange.value() / 1000.0)


from scipy import optimize
import numpy as np
import warnings
import tqdm

from trait2d.analysis.models import ModelLinear, ModelPower

def delete_msd_analysis_results(self):
    """Delete the MSD analysis results."""
    self._msd_analysis_results = None

def get_msd_analysis_results(self):
    """Returns the MSD analysis results."""
    return self._msd_analysis_results

def msd_analysis(self, fraction_fit_points: float = 0.25, n_fit_points: int = None, fit_max_time: float = None, initial_guesses = { }, maxfev = 1000, R : float = 0):
    """ Classical Mean Squared Displacement Analysis for single track

    Parameters
    ----------
    fraction_fit_points: float
        Fraction of points to use for fitting if n_fit_points is not specified.
    n_fit_points: int
        Number of points to user for fitting. Will override fraction_fit_points.
    fit_max_time: float
        Maximum time in fit range. Will override fraction_fit_points and n_fit_points.
    initial_guesses: dict
        Dictionary containing initial guesses for the parameters. Keys can be "model1" and "model2".
        All values default to 1.
    maxfev: int
        Maximum function evaluations by scipy.optimize.curve_fit. The fit will fail if this number is exceeded.
    R: float
        Point scanning across the field of view.
    Returns
    -------
    msd_analysis_results: dict
        Dictionary containing all MSD analysis results. Can also be retreived using `Track.get_msd_analysis_results()`.
    """
    p0 = {"model1" : 2 * [None], "model2" : 3*[None]}
    p0.update(initial_guesses)

    # Calculate MSD if this has not been done yet.
    if self._msd is None:
        self.calculate_msd()

    # Number time frames for this track
    N = self._msd.size

    # This is the time array, as the fits will be MSD vs T
    T = self._t[0:-3]

    dt = self._t[1] - self._t[0]

    # Define the number of points to use for fitting
    if fit_max_time is not None:
        n_points = int(np.argwhere(T < fit_max_time)[-1])
    elif n_fit_points is not None:
        n_points = int(n_fit_points)
    else:
        n_points = int(fraction_fit_points * N)

    # Asserting that the n_fit_points is valid
    assert n_points >= 2, f"n_fit_points={n_points} is not enough"
    if n_points > int(0.25 * N):
        warnings.warn(
            "Using too many points for the fit means including points which have higher measurment errors.")
        # Selecting more points than 25% should be possible, but not advised

    from trait2d.analysis.models import ModelLinear, ModelPower
    model1 = ModelLinear()
    model2 = ModelPower()

    model1.R = R
    model2.R = R
    model1.dt = dt
    model2.dt = dt

    p0_model1 = [0.0, 0.0]
    for i in range(len(p0_model1)):
        if not p0["model1"][i] is None:
            p0_model1[i] = p0["model1"][i]

    reg1 = optimize.curve_fit(
        model1, T[0:n_points], self._msd[0:n_points], p0 = p0_model1, sigma=self._msd_error[0:n_points], maxfev=maxfev, method='trf', bounds=(0.0, np.inf))

    p0_model2 = [0.0, 0.0, 0.0]
    for i in range(len(p0_model2)):
        if not p0["model2"][i] is None:
            p0_model2[i] = p0["model2"][i]
    reg2 = optimize.curve_fit(model2, T[0:n_points], self._msd[0:n_points], p0 = p0_model2, sigma=self._msd_error[0:n_points], maxfev=maxfev, method='trf', bounds=(0.0, np.inf))


    # Compute standard deviation of parameters
    perr_m1 = np.sqrt(np.diag(reg1[1]))
    perr_m2 = np.sqrt(np.diag(reg2[1]))

    # Compute BIC for both models
    m1 = model1(T, *reg1[0])
    m2 = model2(T, *reg2[0])
    bic1 = BIC(m1[0:n_points], self._msd[0:n_points], 2, 1)
    bic2 = BIC(m2[0:n_points], self._msd[0:n_points], 2, 1)
    # FIXME: numerical instabilities due to low position values. should normalize before analysis, and then report those adimentional values.

    # Relative Likelihood for each model
    rel_likelihood_1 = np.exp((-bic1 + min([bic1, bic2])) * 0.5)
    rel_likelihood_2 = np.exp((-bic2 + min([bic1, bic2])) * 0.5)

    self._msd_analysis_results = {}
    self._msd_analysis_results["fit_results"] = {"model1": {"params": reg1[0], "errors" : perr_m1, "bic": bic1, "rel_likelihood": rel_likelihood_1},
                                                "model2": {"params": reg2[0], "errors" : perr_m2, "bic": bic2, "rel_likelihood": rel_likelihood_2}}
    self._msd_analysis_results["n_points"] = n_points

    return self._msd_analysis_results

def plot_msd_analysis_results(self, scale: str = 'log'):
    """Plot the MSD analysis results.

    Parameters
    ----------
    scale: str
        How to scale the plot over time. Possible values: 'log', 'linear'.

    Raises
    ------
    ValueError
        Track has not been analyzed using MSD analysis yet.
    """
    import matplotlib.pyplot as plt
    if self.get_msd_analysis_results() is None:
        raise ValueError(
            "Track as not been analyzed using msd_analysis yet!")

    # Definining the models used for the fit
    model1 = ModelLinear()
    model2 = ModelPower()

    results = self.get_msd_analysis_results()["fit_results"]
    N = self._x.size
    T = self._t[0:-3]
    n_points = self.get_msd_analysis_results()["n_points"]
    reg1 = results["model1"]["params"]
    reg2 = results["model2"]["params"]
    m1 = model1(T, *reg1)
    m2 = model2(T, *reg2)
    rel_likelihood_1 = results["model1"]["rel_likelihood"]
    rel_likelihood_2 = results["model2"]["rel_likelihood"]
    # Plot the results
    plt.figure()
    plt.grid(linestyle='dashed', color='grey')
    if scale == 'linear':
        plt.plot(T, self._msd, color='black', label="Data")
    elif scale == 'log':
        plt.semilogx(T, self._msd, color='black', label="Data")
    else:
        raise ValueError("scale must be 'log' or 'linear'.")
    plt.fill_between(T, self._msd-self._msd_error, self._msd+self._msd_error, color='black', alpha=0.5)
    plt.plot(T[0:n_points], m1[0:n_points],
                label=f"Model1, Rel_Likelihood={rel_likelihood_1:.2e}")
    plt.plot(T[0:n_points], m2[0:n_points],
                label=f"Model2, Rel_Likelihood={rel_likelihood_2:.2e}")
    plt.axvspan(T[0], T[n_points], alpha=0.5,
                color='gray', label="Fitted data")
    plt.xlabel("Time")
    plt.ylabel("MSD")
    plt.legend()
    plt.show()

def MSD_loop(i, pos_x, pos_y, N):
    idx_0 = np.arange(1, N-i-1, 1)
    idx_t = idx_0 + i
    this_msd = (pos_x[idx_t] - pos_x[idx_0])**2 + \
        (pos_y[idx_t] - pos_y[idx_0])**2

    MSD = np.mean(this_msd)
    MSD_error = np.std(this_msd) / np.sqrt(len(this_msd))

    return MSD, MSD_error


def rayleighPDF(x, sigma):
    return x / sigma**2 * np.exp(- x**2 / (2 * sigma**2))


def BIC(pred: list, target: list, k: int, n: int):
    """Bayesian Information Criterion
    Parameters
    ----------
    pred: list
        Model prediction
    target: list
        Model targe

    k: int
        Number of free parameters in the fit
    n: int
        Number of data points used to fit the model
    Returns
    -------
    bic : float
        Bayesian Information Criterion
    """
    # Compute RSS
    RSS = np.sum((np.array(pred) - np.array(target)) ** 2)
    bic = k * np.log(n) + n * np.log(RSS / n)
    return bic


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
GUI iSCAT simulator

@author: mariaa, joel
"""

import matplotlib
matplotlib.use('TkAgg') # This is a bug fix in order to use the GUI on Mac

from trait2d.simulators import HoppingDiffusion, iscat_movie, BrownianDiffusion
import matplotlib.pyplot as plt
import numpy as np
import tkinter as tk
from tkinter import filedialog
from tkinter import ttk

# for plotting
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
# default Matplotlib key bindings

# Global Variables
HOPPING_DIFFUSION = 0  # Label for the hopping diffusion
FREE_DIFFUSION = 1  # Label for the free diffusion

class MainVisual(tk.Frame):
    '''
    class of GUI for simulator
    '''

    def __init__(self, master):

        #define a window
        tk.Frame.__init__(self, master)
        self.master = master
        master.title("TRAIT Simulator")
        master.configure(background='gray')
        # get the monitor size
        self.dpi = 100
        self.monitor_width = master.winfo_screenwidth()
        self.monitor_height = master.winfo_screenheight()
        self.button_size = int(self.monitor_width/25)
        self.fig_size = (int(self.monitor_height/2/self.dpi), int(self.monitor_height/2/self.dpi))
        
        master.protocol('WM_DELETE_WINDOW', self.close_app)

        # # # parameters # # #
        # trajectory generation
        self.Tmax = 10  # Maximal total length [s]
        self.dt = 5e-3  #Time step [s]
        self.L = 10e-6  #Sandbox length [m]
        self.dL = 20e-9  #Compartment map pixel size [m]
        self.Df = 8e-13  # Free diffusion coefficient [m^2/s]
        self.HL = 40e-9  # Average compartment diameter/length [m]
        self.HP = 0.01  #Hopping probability [0-1]
        self.seed = None  #Random generator seed (nonnegative integer)

        self.dynamics_switch = 0  # dynamics type: 0 - diffusion 1 - hopping diffusion
        
        self.trajectory_file_type = "csv"  # file type to save the trajectories
        self.trajectory = {}  # data of the tracks prepared for the csv file
        
        # image generation
        self.resolution = 1e-8
        self.dt_image = 5e-3
        self.snr = 25
        self.background = 0.3
        self.noise_gaussian = 0.15
        self.noise_poisson = True
        self.ratio = "square"
        
        # trajectory generator
        self.TG = HoppingDiffusion(Tmax=self.Tmax, dt=self.dt, L=self.L, dL=self.dL,
                                   Df=self.Df, HL=self.HL, HP=self.HP, seed=self.seed)
        
        # image generator
        self.IG = iscat_movie(tracks=None, resolution=self.resolution, dt=self.dt_image,
                              snr=self.snr, background=self.background,
                              noise_gaussian=self.noise_gaussian,
                              noise_poisson=self.noise_poisson, ratio=self.ratio)
        
     # # # # # # menu to choose files and set tracker parameters # # # # # #

        # name the section

        lbl1 = tk.Label(master=root, text=" TRAJECTORY ", bg='gray', compound=tk.LEFT)
        lbl1.grid(row=0, column=1, columnspan=1, pady=5)

        # separator - line 
        self._separator = ttk.Separator(master=root, orient=tk.VERTICAL)
        self._separator.grid(row=1, column=1, columnspan=8, pady=5, sticky="ew")

        # type of dynamics
        var = tk.IntVar()  # the switch variable

        # variable update
        def update_switch():
            self.dynamics_switch = var.get()

        # dynamics type switch: # 1 - diffusion 0 - hopping diffusion
        self.R1 = tk.Radiobutton(root, text=" Free Diffusion ", variable=var,
                                 value=1, bg='gray', command=update_switch)
        self.R1.grid(row=2, column=1, columnspan=2)

        self.R2 = tk.Radiobutton(root, text=" Hopping Diffusion ", variable=var,
                                 value=0, bg='gray', command=update_switch) #  command=sel)
        self.R2.grid(row=2, column=3, columnspan=3, pady=5)
        
        # setting trajectory parameters
        lbl1 = tk.Label(master=root, text=" Parameters: ", width=30, bg='gray')
        lbl1.grid(row=3, column=1, columnspan=7, pady=5)


        lbl3 = tk.Label(master=root, text="Maximal total length [Tmax, s]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl3.grid(row=4, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.Tmax))
        self.param_Tmax = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_Tmax.grid(row=4, column=3, columnspan=6)

        lbl4 = tk.Label(master=root, text="Time step [dt, s]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl4.grid(row=5, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.dt))
        self.param_dt = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_dt.grid(row=5, column=3, columnspan=6)

        lbl5 = tk.Label(master=root, text=" Sandbox length [L, m]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl5.grid(row=6, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.L))
        self.param_L = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_L.grid(row=6, column=3, columnspan=6)

        lbl2 = tk.Label(master=root, text=" Compartment map pixel size [dL, m]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl2.grid(row=7, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.dL))
        self.param_dL = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_dL.grid(row=7, column=3, columnspan=6)


        lbl6 = tk.Label(master=root, text=" Free diffusion coefficient [Df, "+r'm^2/s]',
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl6.grid(row=8, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.Df))
        self.param_Df = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_Df.grid(row=8, column=3, columnspan=6)

        lbl6 = tk.Label(master=root, text=" Average compartment diameter/length [HL, m]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl6.grid(row=9, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.HL))
        self.param_HL = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_HL.grid(row=9, column=3, columnspan=6)
        
        lbl6 = tk.Label(master=root, text=" Hopping probability [HP, [0,1]]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl6.grid(row=10, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.HP))
        self.param_HP = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_HP.grid(row=10, column=3, columnspan=6)

        lbl6 = tk.Label(master=root, text=" Random generator seed [seed, integer]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl6.grid(row=11, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.seed))
        self.param_seed = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_seed.grid(row=11, column=3, columnspan=6)

        #generate button
        self.button2 = tk.Button(text="    GENERATE    ", command=self.generate_trajectory,
                                 width=int(self.button_size/4), bg='gray') #, height=30)
        self.button2.grid(row=12, column=1, columnspan=2,pady=5)

        # button to save results
        self.button2 = tk.Button(text="    SAVE   ", command=self.save_trajectory,
                                 width=int(self.button_size/4), bg='gray')
        self.button2.grid(row=12, column=3, columnspan=6, pady=5)
        
        #preview button
        self.button2 = tk.Button(text="  SHOW TRACK  ", command=self.show_trajectory,
                                 width=int(self.button_size/4), bg='gray') #, height=30)
        self.button2.grid(row=13, column=1, columnspan=2,pady=5)
        
        # choose the file type
        # type of dynamics
        var_2 = tk.StringVar()  # the switch variable
        var_2.set("csv")
        # variable update
        def update_switch_2():
            self.trajectory_file_type=var_2.get()

        
        # trajectory file type
        self.F1 = tk.Radiobutton(root, text=" csv", variable=var_2,
                                 value="csv", bg='gray',command = update_switch_2 ) #  command=sel)
        self.F1.grid(row=13, column=4, columnspan=1, pady=5)     

        self.F2 = tk.Radiobutton(root, text=" json", variable=var_2,
                                 value="json", bg='gray',command = update_switch_2 ) #  command=sel)
        self.F2.grid(row=13, column=5, columnspan=1, pady=5)  

        self.F3 = tk.Radiobutton(root, text=" pcl", variable=var_2,
                                 value="pcl", bg='gray',command = update_switch_2 ) #  command=sel)
        self.F3.grid(row=13, column=6, columnspan=1, pady=5)     
##################################################

        # separator - line 
        self._separator = ttk.Separator(master=root, orient=tk.VERTICAL)
        self._separator.grid(row=20, column=1, columnspan=8, pady=5, sticky="ew")        
                # name the section

        lbl1 = tk.Label(master=root, text=" IMAGE SEQUENCE ", bg='gray', compound=tk.LEFT)
        lbl1.grid(row=21, column=1, columnspan=1, pady=5)

        # separator - line 
        self._separator = ttk.Separator(master=root, orient=tk.VERTICAL)
        self._separator.grid(row=22, column=1, columnspan=8, pady=5, sticky="ew")
        
        # button to load trajectory
        self.button2 = tk.Button(text="    load trajectory    ", command=self.load_trajectory,
                                 width=int(self.button_size/4), bg='gray')
        self.button2.grid(row=23, column=3, columnspan=6, pady=5)
        
        # setting image parameters
        lbl1 = tk.Label(master=root, text="Parameters: ", width=int(self.button_size/1.5), bg='gray')
        lbl1.grid(row=24, column=1, columnspan=7, pady=5)


        lbl3 = tk.Label(master=root, text="Resolution [resolution, m/pix]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl3.grid(row=25, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.resolution))
        self.param_resolution = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_resolution.grid(row=25, column=3, columnspan=6)

        lbl4 = tk.Label(master=root, text=" Signal to noise ratio [snr, ratio]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl4.grid(row=26, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.snr))
        self.param_snr = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_snr.grid(row=26, column=3, columnspan=6)

        lbl5 = tk.Label(master=root, text=" Background intensity [background, [0,1]]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl5.grid(row=27, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.background))
        self.param_background = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_background.grid(row=27, column=3, columnspan=6)

        lbl2 = tk.Label(master=root, text=" Gaussian noise variance [noise_gaussian]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl2.grid(row=28, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.noise_gaussian))
        self.param_noise_gaussian = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_noise_gaussian.grid(row=28, column=3, columnspan=6)


        lbl6 = tk.Label(master=root, text=" Poisson noise variance [noise_poisson, bool]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl6.grid(row=29, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.noise_poisson))
        self.param_noise_poisson = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_noise_poisson.grid(row=29, column=3, columnspan=6)

        lbl2 = tk.Label(master=root, text=" Temporal resolution [dt_image, frame/sec]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl2.grid(row=30, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.dt_image))
        self.param_dt_image = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_dt_image.grid(row=30, column=3, columnspan=6)


        lbl6 = tk.Label(master=root, text=" Ratio [ratio]",
                        width=self.button_size, bg='gray', compound=tk.LEFT)
        lbl6.grid(row=31, column=1, columnspan=2)
        v = tk.StringVar(root, value=str(self.ratio))
        self.param_ratio = tk.Entry(root, width=int(self.button_size/4), text=v)
        self.param_ratio.grid(row=31, column=3, columnspan=6)
        
        #preview button
        self.button2 = tk.Button(text=" load psf", command=self.load_psf,
                                 width=int(self.button_size/4), bg='gray') #, height=30)
        self.button2.grid(row=32, column=1, columnspan=2,pady=5)
        
 
        #preview button
        self.button2 = tk.Button(text=" GENERATE and SHOW ", command=self.generate_images,
                                 width=int(self.button_size/4), bg='gray') #, height=30)
        self.button2.grid(row=33, column=1, columnspan=2,pady=5)

        # button to run the tracker and save results
        self.button2 = tk.Button(text="    SAVE   ", command=self.save_images,
                                 width=int(self.button_size/4), bg='gray')
        self.button2.grid(row=33, column=3, columnspan=6, pady=10, padx=30)

    def generate_trajectory(self):
        '''
        function to generate trajectory
        '''
        # update the parameters        
        self.read_parameters()
        
        if self.dynamics_switch == HOPPING_DIFFUSION:
            self.TG = HoppingDiffusion(Tmax=self.Tmax, dt=self.dt, L=self.L, dL=self.dL,
                                       Df=self.Df, HL=self.HL, HP=self.HP, seed=self.seed)
            self.TG.run()
        elif self.dynamics_switch == FREE_DIFFUSION:
            self.TG = BrownianDiffusion(Tmax=self.Tmax, dt=self.dt, L=self.L, dL=self.dL,
                                        d=self.Df,  seed=self.seed)
            self.TG.run()            
                 
        # print("generate_trajectory(self)") ## Debug print
        self.trajectory = self.TG.trajectory

    def save_trajectory(self):
        '''
        save trajectory
        '''
        # update the parameters        
        self.read_parameters()
        # select file location and name
        save_file = tk.filedialog.asksaveasfilename()
        if not(save_file.endswith(self.trajectory_file_type)):
                save_file += "."+self.trajectory_file_type
                
        self.TG.save_trajectory(save_file, self.trajectory_file_type)

         
        print("save trajectory into a file: ", save_file)
        
    def show_trajectory(self):
        '''
        plot trajectory in separate window
        '''
        # Update the parameters
        self.read_parameters()

        # DrawingArea
        novi = tk.Toplevel()
        novi.title("trajectory plot")

        if self.dynamics_switch == HOPPING_DIFFUSION:
            title = "Hopping diffusion"
        elif self.dynamics_switch == FREE_DIFFUSION:
            title = "Free diffusion"
        else:
            title = "Diffusion"
        
        fig = plt.figure(figsize=self.fig_size)
        self.TG.display_trajectory(time_resolution=0.5e-3, limit_fov=False, alpha=0.8,
                                   title=title)
        plt.xlabel("x")
        plt.ylabel("y")            
                # DrawingArea
        canvas = FigureCanvasTkAgg(fig, master=novi)
        canvas.draw()
        canvas.get_tk_widget().grid(row=0, column=0)

        # print("trajectory is plotted") # DEBUG print

    def read_parameters(self):
        '''
        read parameters from the GUI
        '''
        # trajectory
        if self.param_Tmax.get()!='':
            self.Tmax=float(self.param_Tmax.get())

        if self.param_dt.get()!='':
            self.dt=float(self.param_dt.get())

        if self.param_L.get()!='':
            self.L=float(self.param_L.get())

        if self.param_dL.get()!='':
            self.dL=float(self.param_dL.get())

        if self.param_Df.get()!='':
            self.Df=float(self.param_Df.get())

        if self.param_HL.get()!='':
            self.HL=float(self.param_HL.get())

        if self.param_HP.get()!='':
            self.HP=float(self.param_HP.get())

        if self.param_seed.get()!='':
            self.seed=int(self.param_seed.get())
        else:
            self.seed = None # To use a random seed
            
        # image sequence

        if self.param_resolution.get()!='':
            self.resolution=float(self.param_resolution.get())

        if self.param_dt_image.get()!='':
            self.dt_image=float(self.param_dt_image.get())

        if self.param_snr.get()!='':
            self.snr=float(self.param_snr.get())

        if self.param_background.get()!='':
            self.background=float(self.param_background.get())

        if self.param_noise_gaussian.get()!='':
            self.noise_gaussian=float(self.param_noise_gaussian.get())

        if self.param_noise_poisson.get()!='':
            self.noise_poisson=bool(self.param_noise_poisson.get())

        if self.param_ratio.get()!='':
            self.ratio=str(self.param_ratio.get())
            

    def load_trajectory(self):
        '''
        function to load trajectory
        '''
        
        print("load_trajectory(self)")

        filename = tk.filedialog.askopenfilename()
        root.update()   
        
        # load from csv file
        
        self.IG.load_tracks(filename, field_x="x", field_y="y", field_t="t", field_id="id")
        
        
        self.trajectory=self.IG.tracks
        

    def generate_images(self):
        '''
        function to generate image sequence
        '''
        # update the parameters        
        self.read_parameters()
        
        # define image generator class with new parameters and the trajectory
        self.IG.tracks=self.trajectory
        self.IG.resolution=self.resolution
        self.IG.dt=self.dt_image
        self.IG.snr=self.snr
        self.IG.background=self.background
        self.IG.noise_gaussian=self.noise_gaussian
        self.IG.noise_poisson=self.noise_poisson
        self.IG.ratio=self.ratio
        
        # run the generator
        self.IG.run()
        
        # plot the average image
        
        img = np.average(self.IG.movie, axis=0)
        
        # DrawingArea
        novi = tk.Toplevel()
        novi.title("Projection of the image sequence")
        
        fig = plt.figure(figsize=self.fig_size)

#        time_interval = int(np.ceil(0.5e-3 / self.dt))
#        x = self.trajectory["x"][0::time_interval]
#        y = self.trajectory["y"][0::time_interval]
#        plt.plot(x, y, alpha=0.8)
        plt.imshow(img.T, origin='lower', cmap="gray")
        plt.xlabel("x")
        plt.ylabel("y")
        plt.title("Projection of the image sequence")
            
                # DrawingArea
        canvas = FigureCanvasTkAgg(fig, master=novi)
        canvas.draw()
        canvas.get_tk_widget().grid(row=0, column=0)      
        
        print("image is generated")

    def save_images(self):
        '''
        function to save image sequence
        '''
        # select file location and name
        save_file_name = tk.filedialog.asksaveasfilename()
        if not(save_file_name.endswith("tiff")) and not(save_file_name.endswith("tif")):
            save_file_name += ".tiff"

        self.IG.save(save_file_name)
        print(f"save images: {save_file_name}")
        
    def load_psf(self):
        '''
        load psf
        '''
        filename = tk.filedialog.askopenfilename()
        root.update() 
        
        self.IG.load_psf(filename)

    def close_app(self):
        '''
        quit all the proccesses while closing the GUI
        '''
        self.quit()

def main():
    global root
    root = tk.Tk()
    MainVisual(root)
    root.mainloop()

if __name__ == "__main__":
    main()
